#ifndef __MODEL_MGR_H__
#define __MODEL_MGR_H__

#include <stdint.h>
#include "cmsis_os2.h"
#include "ipc.h"             /* for MULTI_MODEL_MAX */
#include "base.h"
#include "model_type.h"
#include "jpeg_intf.h"

/* a term means ALL_MODELS */
#define KMDW_MODEL_ALL_MODELS  -1

/* flash memory table
 *    BOOT related       :  192K(reserved)      0x0         - 0x2FFFF
 *    FLASH_ADDR_HEAD    :                      0x30000
 *    FID database       :  512K(reserved)      0x80000     - 0xFFFFF
 *    fw_info.bin        :    4K(reserved)      0x100000    - 0x100FFF  : real use: depend on model count
 *      - model count : 4 bytes
 *      - models      :sizeof(struct kdp_model_s)*count
 *      - ddr end addr:4 bytes
 *
 *    bin files for models: (4K aligned)        0x101000     - ??       : real use: depend model size
 *      - cmd.bin + weight.bin + setup.bin
 *    FLASH_ADDR_END                            0x101000 + model size   -> + 4K alignment
 */
/* MAX model count for DME and flash */
#define KMDW_MODEL_MAX_MODEL_COUNT MULTI_MODEL_MAX

/* These crc values will ONLY change if model is changed in FLASH */
#define MODEL_INFO_CRC16_VALUE  0x5341
#define MODEL_INFO_SUM32_VALUE  0x00eee074

#ifndef FLASH_MODEL_HAS_CRC_CHECK
#define MODEL0_FD_CRC16_VALUE   0x939f   /* original: 0x4318 */
#define MODEL1_LM_CRC16_VALUE   0x1cd0   /* original: 0xfce0 */
#define MODEL3_FR_CRC16_VALUE   0x5df0   /* original: 0xb359 */
#define MODEL0_FD_SUM32_VALUE   0xf9e571d8
#define MODEL1_LM_SUM32_VALUE   0x3f96c4fc
#define MODEL3_FR_SUM32_VALUE   0xbc998f45
#endif

enum kmdw_model_rc {
    // 0 - 9 is reserved for ncpu return
    // defined in ipc.h
    // IMAGE_STATE_INACTIVE == 0
    // IMAGE_STATE_ACTIVE == 1
    // IMAGE_STATE_NPU_DONE == 2
    // IMAGE_STATE_DONE == 3
    KMDW_MODEL_RUN_RC_ABORT = 10,
    KMDW_MODEL_RUN_RC_ERROR = 11,
    KMDW_MODEL_RUN_RC_END
};

typedef struct kmdw_model_run_time_s {
    float    round_trip_time;
    float    pre_proc_time;
    float    npu_proc_time;
    float    post_proc_time;
    float    rslt_pop_interval;
    float    scpu_send_ipc_ack;
    float    ncpu_req_ipc_ack;
    float    ncpu_rslt_ipc_ack;
    float    ncpu_img_req_to_rcv_new_img;
} kmdw_model_run_time_t;

//a mask for easy read fw_info data
typedef struct kmdw_model_fw_info_s {
    uint32_t                model_count;
    struct kdp_model_s      models[1];   // just maintain the address of dynamical count of models
} kmdw_model_fw_info_t;

//a mask for easy read fw_info data
typedef struct kmdw_model_fw_info_ext_s {
    uint32_t                model_dram_addr_end;
    uint32_t                model_total_size;
    uint32_t                model_checksum;
} kmdw_model_fw_info_ext_t;


/* ############################
 * ##    Public Functions    ##
 * ############################ */

/**
 * @brief Init model functionality
 */
void kmdw_model_init(void);

/**
 * @brief A wrapper of load_model from flash
 * @param model_index_p: model info index,
 *                    0-n: info_index of model to load
 *                    -1 means to load all models
 * @return 0: failes;  1: OK(means 1 model is loaded)
 */
int32_t kmdw_model_load_model(int8_t model_info_index_p);


/**
 * @brief A wrapper of load_model_info
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 * @return reloaded model count; 0 means failed
 */
int32_t kmdw_model_reload_model_info(bool from_ddr);


/**
 * @brief Refresh all models
 * @return refreshed model count; 0 means failed
 */
int32_t kmdw_model_refresh_models(void);


/**
 * @brief Output model_info of specified index
 * @param[in] idx_p the index of programmed models
 * @return model_info defined in ipc.h
 */
struct kdp_model_s* kmdw_model_get_model_info(int idx_p);


/**
 * @brief Output total number of models and all model IDs
 * @param[in] from_ddr: if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return a list of model count, id0, id1, id2
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t *kmdw_model_get_all_model_info(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Output model end address defined in model file
 * @param[in] from_ddr if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return model end address when loaded
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t kmdw_model_get_model_end_addr(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Output crc value
 * @param[in] from_ddr: if models are in ddr or flash
 * @param[in] trust_ddr_data trust ddr data
 * @return crc value
 * @note about trust_ddr_data, ddr data might be dirty after soft/hw reset
 *       trust_ddr_data means no need to check validness of ddr data
 */
uint32_t kmdw_model_get_crc(bool from_ddr, bool trust_ddr_data);

/**
 * @brief Get the buffer address of fw_info
 * @return the buffer address of fw_info
 */
kmdw_model_fw_info_t* kmdw_model_get_fw_info_buf(void);


/**
 * @brief Check if a model is in flash
 * @param[in] model_type model type
 * @return 1 - the model is in flash; 0 - not in flash
 */
int kmdw_model_is_model_in_flash(uint32_t model_type);


/**
 * @brief Specify output address for model run in ncpu/npu
 * @return always 0
 * @note must be called after kmdw_model_config_model()
 */
int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag);


/**
 * @brief Config model image
 * @param img_cfg image config
 * @param crop_box image crop config
 * @param pad_values image padding config
 * @param ext_param extra param
 */
void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param);


/**
 * @brief Get raw image config
 * @param idx image index
 * @return raw image config
 */
struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx);


/**
 * @brief Run model
 * @param tag model tag
 * @param output model output
 * @param model_type model type
 * @param dme DME mode
 * @return kmdw_model_rc
 */
int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool dme);


/**
 * @brief Abort model execution
 */
void kmdw_model_abort(void);


/**
 * @brief Get round-trip/pre/npu/post times
 * @param [in]  img_idx raw image index
 * @param [out] run_time run time of pre/npu/post/
 * @return void
 */
void kmdw_model_get_run_time(int img_idx, kmdw_model_run_time_t *run_time);


/**
 * @brief Check if a model is loaded in memory
 * @param[in] model_type model type
 * @return 1 - loaded in memory; 0 - not loaded
 */
int kmdw_model_is_model_loaded(uint32_t model_type);











void model_jpeg_enc_cfg(uint32_t img_seq, uint32_t jpeg_in_buf_addr, uint32_t jpeg_in_size, uint32_t jpeg_out_buf_addr,
                        uint32_t exec_rslt_addr, uint32_t w, uint32_t h, uint32_t quality, uint32_t fmt, osEventFlagsId_t *pEvt, uint32_t evt_flag);

void model_jpeg_enc_start_run(void);
jpeg_encode_result_t* model_jpeg_enc_get_rslt(void);

void model_jpeg_dec_cfg(uint32_t img_seq, uint32_t jpeg_in_buf_addr, uint32_t jpeg_in_size, uint32_t jpeg_out_buf_addr,
                        uint32_t exec_rslt_addr, uint32_t w, uint32_t h, uint32_t fmt, osEventFlagsId_t *pEvt, uint32_t evt_flag);

void model_jpeg_dec_start_run(void);
jpeg_decode_result_t* model_jpeg_dec_get_rslt(void);

extern int get_yuv_y_len(jpeg_fmt_t fmt, int w, int h);
extern int get_yuv_u_len(jpeg_fmt_t fmt, int w, int h);
extern int get_yuv_v_len(jpeg_fmt_t fmt, int w, int h);


#if DEBUG

/**
 * @brief Dump model debug info
 */
void kmdw_model_dump_model_info(void);

#endif

#endif
