/*
 * Kneron Application general functions
 *
 * Copyright (C) 2021 Kneron, Inc. All rights reserved.
 *
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "kmdw_console.h"

#include "kmdw_inference_app.h"
#include "kdp2_inf_generic_raw.h"

uint32_t kdp2_get_raw_output_info_size(void)
{
    return sizeof(kdp2_ipc_generic_raw_result_t) + sizeof(_720_raw_cnn_res_t);
}

void kdp2_generic_raw_inference(void *inf_input_buf)
{
    // 'inf_input_buf' and 'result_buf' are provided by kdp2 middleware
    // the content of 'inf_input_buf' is transmitted from host SW = header + image
    // 'result_buf' is used to carry inference result back to host SW = header + inferernce result (from ncpu/npu)

    kdp2_ipc_generic_raw_inf_header_t *input_header = (kdp2_ipc_generic_raw_inf_header_t *)inf_input_buf;

    // config image preprocessing and model settings
    kmdw_inference_app_config_t inf_config;
    memset(&inf_config, 0, sizeof(kmdw_inference_app_config_t)); // for safety let default 'bool' to 'false'

    // image buffer address should be just after the header
    inf_config.image_buf = (void *)((uint32_t)input_header + sizeof(kdp2_ipc_generic_raw_inf_header_t));
    inf_config.image_width = input_header->image_header.width;
    inf_config.image_height = input_header->image_header.height;
    inf_config.image_channel = (input_header->image_header.image_format == KP_IMAGE_FORMAT_RAW8) ? 1 : 3;
    inf_config.image_format = input_header->image_header.image_format;
    inf_config.image_norm = input_header->image_header.normalize_mode;
    inf_config.model_id = input_header->image_header.model_id;
    inf_config.enable_crop = (0 < input_header->image_header.crop_count);      // use crop
    inf_config.enable_raw_output = true;                                       // raw output no post-processing
    inf_config.image_resize = input_header->image_header.resize_mode;          // user's choice
    inf_config.image_padding = input_header->image_header.padding_mode;        // user's choice
    // inf_config.enable_parallel = false;                                          // due to no post-processing

    // need to know model raw output size for result transfer size
    uint32_t model_raw_out_size = kmdw_inference_app_get_model_raw_output_size(inf_config.model_id);

    int crop_count = input_header->image_header.crop_count;

    if (false == inf_config.enable_crop) {
        // now get an available free result buffer
        // normally the begin part of result buffer should contain app-defined result header
        // and the rest is for ncpu/npu inference output data
        int output_header_buf_size;
        void *result_buf = kmdw_inference_app_result_get_free_buffer(&output_header_buf_size);

        // leave some space for result header
        void *ncpu_result_buf = (void *)((uint32_t)result_buf + sizeof(kdp2_ipc_generic_raw_result_t));

        inf_config.ncpu_result_buf = ncpu_result_buf; // give result buffer for ncpu/npu

        inf_config.result_callback = NULL;
        inf_config.pad_value = NULL;
        inf_config.user_define_data = NULL;

        // run preprocessing and inference, trigger ncpu/npu to do the work
        int ret = kmdw_inference_app_execute(&inf_config);

        kdp2_ipc_generic_raw_result_t *output_header = (kdp2_ipc_generic_raw_result_t *)result_buf;

        // header_stamp is a must to correctly transfer result data back to host SW
        output_header->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
        output_header->header_stamp.job_id = KDP2_INF_ID_GENERIC_RAW;
        output_header->inf_number = input_header->image_header.inference_number;         // sync the inference number
        output_header->crop_number = 0;                                                  // sync the crop number
        output_header->is_last_crop = 1;

        if (ret == KP_SUCCESS) {
            output_header->header_stamp.status_code = KP_SUCCESS;
            output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_result_t) + sizeof(_720_raw_cnn_res_t) + model_raw_out_size;
        } else {
            // some sort of inference error
            output_header->header_stamp.status_code = ret;
            output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_result_t);
        }

        kmdw_inference_app_result_enqueue((void *)output_header, output_header_buf_size, false);
        return;
    }

    // remember: one crop, one inference result !
    for (int c = 0; c < crop_count; c++)
    {
        // now get an available free result buffer
        // normally the begin part of result buffer should contain app-defined result header
        // and the rest is for ncpu/npu inference output data
        int output_header_buf_size;
        void *result_buf = kmdw_inference_app_result_get_free_buffer(&output_header_buf_size);

        // leave some space for result header
        void *ncpu_result_buf = (void *)((uint32_t)result_buf + sizeof(kdp2_ipc_generic_raw_result_t));

        inf_config.crop_area = input_header->image_header.inf_crop[c];
        inf_config.ncpu_result_buf = ncpu_result_buf; // give result buffer for ncpu/npu

        inf_config.result_callback = NULL;
        inf_config.pad_value = NULL;
        inf_config.user_define_data = NULL;

        // run preprocessing and inference, trigger ncpu/npu to do the work
        int ret = kmdw_inference_app_execute(&inf_config);

        kdp2_ipc_generic_raw_result_t *output_header = (kdp2_ipc_generic_raw_result_t *)result_buf;

        // header_stamp is a must to correctly transfer result data back to host SW
        output_header->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
        output_header->header_stamp.job_id = KDP2_INF_ID_GENERIC_RAW;
        output_header->inf_number = input_header->image_header.inference_number;         // sync the inference number
        output_header->crop_number = input_header->image_header.inf_crop[c].crop_number; // sync the crop number
        output_header->is_last_crop = (c == crop_count - 1) ? 1 : 0;

        if (ret == KP_SUCCESS)
        {
            output_header->header_stamp.status_code = KP_SUCCESS;
            output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_result_t) + sizeof(_720_raw_cnn_res_t) + model_raw_out_size;
        }
        else
        {
            // some sort of inference error
            output_header->header_stamp.status_code = ret;
            output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_result_t);
        }

        kmdw_inference_app_result_enqueue((void *)output_header, output_header_buf_size, false);
    }
}

void kdp2_generic_raw_inference_bypass_pre_proc(void *inf_input_buf)
{
    // 'inf_input_buf' and 'result_buf' are provided by kdp2 middleware
    // the content of 'inf_input_buf' is transmitted from host SW = header + image
    // 'result_buf' is used to carry inference result back to host SW = header + inferernce result (from ncpu/npu)

    kdp2_ipc_generic_raw_inf_bypass_pre_proc_header_t *input_header = (kdp2_ipc_generic_raw_inf_bypass_pre_proc_header_t *)inf_input_buf;

    // config image preprocessing and model settings
    kmdw_inference_app_config_t inf_config;
    memset(&inf_config, 0, sizeof(kmdw_inference_app_config_t)); // for safety let default 'bool' to 'false'

    // image buffer address should be just after the header
    inf_config.image_buf = (void *)((uint32_t)input_header + sizeof(kdp2_ipc_generic_raw_inf_bypass_pre_proc_header_t));
    inf_config.image_buf_size = input_header->image_header.image_buffer_size;
    inf_config.bypass_pre_proc = true;
    inf_config.model_id = input_header->image_header.model_id;
    inf_config.enable_raw_output = true;                                       // raw output no post-processing

    // need to know model raw output size for result transfer size
    uint32_t model_raw_out_size = kmdw_inference_app_get_model_raw_output_size(inf_config.model_id);

    // now get an available free result buffer
    // normally the begin part of result buffer should contain app-defined result header
    // and the rest is for ncpu/npu inference output data
    int output_header_buf_size;
    void *result_buf = kmdw_inference_app_result_get_free_buffer(&output_header_buf_size);

    // leave some space for result header
    void *ncpu_result_buf = (void *)((uint32_t)result_buf + sizeof(kdp2_ipc_generic_raw_bypass_pre_proc_result_t));

    inf_config.ncpu_result_buf = ncpu_result_buf; // give result buffer for ncpu/npu
    inf_config.result_callback = NULL;
    inf_config.pad_value = NULL;
    inf_config.user_define_data = NULL;

    // run preprocessing and inference, trigger ncpu/npu to do the work
    int ret = kmdw_inference_app_execute(&inf_config);

    kdp2_ipc_generic_raw_bypass_pre_proc_result_t *output_header = (kdp2_ipc_generic_raw_bypass_pre_proc_result_t *)result_buf;

    // header_stamp is a must to correctly transfer result data back to host SW
    output_header->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
    output_header->header_stamp.job_id = KDP2_INF_ID_GENERIC_RAW_BYPASS_PRE_PROC;
    output_header->inf_number = input_header->image_header.inference_number;         // sync the inference number
    output_header->crop_number = 0;
    output_header->is_last_crop = 1;

    if (ret == KP_SUCCESS) {
        output_header->header_stamp.status_code = KP_SUCCESS;
        output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_bypass_pre_proc_result_t) + sizeof(_720_raw_cnn_res_t) + model_raw_out_size;
    } else {
        // some sort of inference error
        output_header->header_stamp.status_code = ret;
        output_header->header_stamp.total_size = sizeof(kdp2_ipc_generic_raw_bypass_pre_proc_result_t);
    }

    kmdw_inference_app_result_enqueue((void *)output_header, output_header_buf_size, false);
}
