/*
 * Kneron Application general functions
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

// #define DEBUG_PRINT
#define ENABLE_INFERENCE_TIMEOUT_WATCH

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "kdrv_gdma3.h"
#include "kmdw_memory.h"
#include "kmdw_model.h"
#include "kmdw_console.h"
#include "membase.h"

#include "kp_struct.h"
#include "kmdw_inference_app.h"
#include "dual_fifo2.h"           /*private functions for kmdw_inference*/
#include "kdp2_inf_generic_raw.h" /*private fucntions for kmdw_inference*/

#ifdef DEBUG_PRINT
#include "kmdw_console.h"
#define dbg_print(__format__, ...) kmdw_printf("[inf]"__format__, ##__VA_ARGS__)
#else
#define dbg_print(__format__, ...)
#endif

static dual_fifo2_t _image_fifioq = 0;
static dual_fifo2_t _result_fifoq = 0;

static osEventFlagsId_t g_result_event;

static kmdw_inference_app_callback_t _app_entry_func = NULL;

static volatile int g_inf_index = 0;
static volatile uint32_t g_num_parallel_inf = 0;
static volatile uint32_t g_num_parallel_result = 0;

typedef struct
{
    void *inf_result_buf;
    int inf_result_buf_size;
    void *ncpu_result_buf;
    kmdw_inference_app_result_callback_t result_callback_func;
} result_context_t;

// CNN_Header and NetInput_Node are from kneron_api_data.h in ncpu firmware to parse setup.bin
typedef struct CNN_Header {
    uint32_t crc;
    uint32_t version;
    uint32_t reamaining_models;
    uint32_t model_type;
    uint32_t application_type;
    uint32_t dram_start;
    uint32_t dram_size;
    uint32_t cmd_start;
    uint32_t cmd_size;
    uint32_t weight_start;
    uint32_t weight_size;
    uint32_t input_start;
    uint32_t input_size;
    uint32_t input_num;
    uint32_t output_num;
} CNN_Header;

// node_id = 5
typedef struct NetInput_Node {
    uint32_t node_id;
    uint32_t input_index;
    uint32_t input_format;
    uint32_t input_row;
    uint32_t input_col;
    uint32_t input_channel;
    uint32_t input_start;
    uint32_t input_size;
    uint32_t input_radix;
} NetInput_Node;

#define MAX_OUTPUT_CONTEXT_NUM 6 // smaller than IPC_IMAGE_MAX
static result_context_t g_result_ctx[MAX_OUTPUT_CONTEXT_NUM] = {0};

extern void kdp2_generic_raw_inference(void *inf_input_buf);
extern void kdp2_generic_raw_inference_bypass_pre_proc(void *inf_input_buf);

void kmdw_inference_image_dispatcher_thread(void *argument)
{
    while (1)
    {
        buffer_object_t fifoq_obj; // fifoq buffer object

        // blocking-wait image buffer object
        dual_fifo2_dequeue_data(_image_fifioq, &fifoq_obj, osWaitForever);

        void *inf_input_buf = (void *)fifoq_obj.buffer_addr; // contains header + image

        dbg_print("got a image buffer for inference: buf 0x%x\n", inf_input_buf);

        kp_inference_header_stamp_t *header_stamp = (kp_inference_header_stamp_t *)inf_input_buf;

        if (header_stamp->job_id == KDP2_INF_ID_GENERIC_RAW)
            kdp2_generic_raw_inference(inf_input_buf);
        else if (header_stamp->job_id == KDP2_INF_ID_GENERIC_RAW_BYPASS_PRE_PROC)
            kdp2_generic_raw_inference_bypass_pre_proc(inf_input_buf);
        else
            _app_entry_func(inf_input_buf);

        // return buffer back to fifoq
        dual_fifo2_put_free_buffer(_image_fifioq, fifoq_obj, osWaitForever);
    }
}

void kmdw_inference_result_handler_callback_thread(void *argument)
{
    uint32_t result_index = 0; // next result sequence index
    uint32_t wait_result_flag;
    int timeout_count = 0;

#define INF_TIMEOUT 2000 // twice

    while (1)
    {
        wait_result_flag = (0x1 << result_index);

        uint32_t flags = osEventFlagsWait(g_result_event, wait_result_flag, osFlagsWaitAny, INF_TIMEOUT);

        dbg_print("result_get: osEventFlagsWait() return 0x%x\n", flags);

        // a tricky way to check if inf timeout
        if (flags == (uint32_t)osFlagsErrorTimeout)
        {
            if (g_num_parallel_inf > g_num_parallel_result)
                timeout_count++;

            if (timeout_count >= 2)
            {
                kmdw_printf("[inf] parallel inference timeout\n");
                kmdw_printf("inf req %d done %d timeout %d secs\n", g_num_parallel_inf, g_num_parallel_result, timeout_count);

                void *inf_result_buf = g_result_ctx[result_index].inf_result_buf;
                int inf_result_buf_size = g_result_ctx[result_index].inf_result_buf_size;
                void *ncpu_result_buf = g_result_ctx[result_index].ncpu_result_buf;
                g_result_ctx[result_index].result_callback_func(KP_FW_INFERENCE_TIMEOUT_103, inf_result_buf, inf_result_buf_size, ncpu_result_buf);

                return; // FIXME, game over ?
            }
        }
        else if (flags == wait_result_flag)
        {
            void *inf_result_buf = g_result_ctx[result_index].inf_result_buf;
            int inf_result_buf_size = g_result_ctx[result_index].inf_result_buf_size;
            void *ncpu_result_buf = g_result_ctx[result_index].ncpu_result_buf;
            g_result_ctx[result_index].result_callback_func(KP_SUCCESS, inf_result_buf, inf_result_buf_size, ncpu_result_buf);

            g_num_parallel_result++;

            if (++result_index >= MAX_OUTPUT_CONTEXT_NUM)
                result_index = 0;

            timeout_count = 0;
        }
        else
        {
            // should not be here
            kmdw_printf("[inf] error flag 0x%x\n", flags);
        }
    }
}

/* ############################
 * ##    public functions    ##
 * ############################ */

int kmdw_inference_app_execute(kmdw_inference_app_config_t *inf_config)
{
    dbg_print("run image inference:\n");

    struct kdp_img_cfg ncpu_img_config;
    struct kdp_crop_box_s crop_box;
    struct kdp_crop_box_s *p_crop_box = NULL; // if enable_crop, it points to 'crop_box'

    ncpu_img_config.image_mem_addr = (uint32_t)inf_config->image_buf;
    ncpu_img_config.image_mem_len = 0; // FIXME ??
    ncpu_img_config.image_col = inf_config->image_width;
    ncpu_img_config.image_row = inf_config->image_height;
    ncpu_img_config.image_ch = inf_config->image_channel;
    ncpu_img_config.image_buf_active_index = g_inf_index;

    dbg_print("image_mem_addr = 0x%x\n", ncpu_img_config.image_mem_addr);
    dbg_print("image_col = %d\n", ncpu_img_config.image_col);
    dbg_print("image_row = %d\n", ncpu_img_config.image_row);
    dbg_print("image_ch = %d\n", ncpu_img_config.image_ch);

    ncpu_img_config.image_format = 0; // sycn with 'ncpu_config'

    // if no post-processing
    if (inf_config->enable_raw_output)
        ncpu_img_config.image_format |= IMAGE_FORMAT_RAW_OUTPUT;

    // enable parallel post-processing
    if (inf_config->enable_parallel)
        ncpu_img_config.image_format |= IMAGE_FORMAT_PARALLEL_PROC;

    // Check if image width, heigth > model width, height
    uint32_t num_models = 0;

    uint32_t *p_all_model_info =
        kmdw_model_get_all_model_info(true /*from ddr*/, false /*trust ddr data*/);
    if (p_all_model_info)
        num_models = p_all_model_info[0];
    else
        return KP_ERROR_MODEL_NOT_LOADED_35;

    struct kdp_model_s *p_model_info = NULL;

    uint32_t model_width = 0;
    uint32_t model_height = 0;

    for (int i = 0; i < num_models; i++)
    {
        p_model_info = kmdw_model_get_model_info(i);

        if (p_model_info)
        {
            if (p_model_info->model_type == inf_config->model_id)
            {
                // find network input node, we observed that it's put right after CNN_Header stored in setup_mem_addr in many models
                // FIXME: network input node might not be the first node put right after CNN_Header stored in setup_mem_addr
                //        see fw_model_parse() and fw_get_netinput_node() in fw_model_parse.h/.c for how to retrieve network input node from setup_mem_addr
                NetInput_Node *net_input_node = (NetInput_Node *)(p_model_info->setup_mem_addr + sizeof(CNN_Header));

                model_width = net_input_node->input_col;
                model_height = net_input_node->input_row;
                break;
            }
        }
    }

    if (model_width == 0 && model_height == 0)
        return KP_ERROR_MODEL_NOT_LOADED_35;

    if (inf_config->bypass_pre_proc == false) // enable pre-processing
    {
        switch (inf_config->image_format)
        {
        case KP_IMAGE_FORMAT_RGB565:
            ncpu_img_config.image_format |= NPU_FORMAT_RGB565;
            break;
        case KP_IMAGE_FORMAT_YUYV:
            ncpu_img_config.image_format |= NPU_FORMAT_YCBCR422_Y1CBY0CR;
            break;
        case KP_IMAGE_FORMAT_RGBA8888:
            ncpu_img_config.image_format |= 0x0D; // FIXME
            break;
        case KP_IMAGE_FORMAT_RAW8:
            ncpu_img_config.image_format |= NPU_FORMAT_NIR;
            break;
        default:
            break;
        }

        switch (inf_config->image_norm)
        {
        case KP_NORMALIZE_DISABLE:
        case KP_NORMALIZE_CUSTOMIZED_DEFAULT:
            break;
        case KP_NORMALIZE_YOLO:
        case KP_NORMALIZE_CUSTOMIZED_DIV2:
            ncpu_img_config.image_format |= IMAGE_FORMAT_RIGHT_SHIFT_ONE_BIT;
            break;
        case KP_NORMALIZE_KNERON:
        case KP_NORMALIZE_TENSOR_FLOW:
        case KP_NORMALIZE_CUSTOMIZED_SUB128:
            ncpu_img_config.image_format |= (uint32_t)IMAGE_FORMAT_SUB128;
            break;
        case KP_NORMALIZE_CUSTOMIZED_SUB128_DIV2:
            ncpu_img_config.image_format |= ((uint32_t)IMAGE_FORMAT_SUB128 | IMAGE_FORMAT_RIGHT_SHIFT_ONE_BIT);
            break;
        }

        if (inf_config->enable_crop)
        {
            crop_box.left = inf_config->crop_area.x1;
            crop_box.top = inf_config->crop_area.y1;
            crop_box.right = inf_config->image_width - inf_config->crop_area.width - inf_config->crop_area.x1;
            crop_box.bottom = inf_config->image_height - inf_config->crop_area.height - inf_config->crop_area.y1;
            p_crop_box = &crop_box;
        }

        if ((KP_RESIZE_DISABLE == inf_config->image_resize) && (KP_PADDING_DISABLE == inf_config->image_padding))
        {
            if (inf_config->enable_crop)
            {
                if ((inf_config->crop_area.width != model_width) ||
                    (inf_config->crop_area.height != model_height))
                {
                    return KP_FW_IMAGE_SIZE_NOT_MATCH_MODEL_INPUT_107;
                }
            }
            else
            {
                if ((inf_config->image_width != model_width) ||
                    (inf_config->image_height != model_height))
                {
                    return KP_FW_IMAGE_SIZE_NOT_MATCH_MODEL_INPUT_107;
                }
            }
        }
        else if ((KP_RESIZE_DISABLE == inf_config->image_resize) && (KP_PADDING_DISABLE != inf_config->image_padding))
        {
            return KP_FW_NOT_SUPPORT_PREPROCESSING_108;
        }
        else if ((KP_RESIZE_DISABLE != inf_config->image_resize) && (KP_PADDING_DISABLE == inf_config->image_padding))
        {
            ncpu_img_config.image_format |= IMAGE_FORMAT_CHANGE_ASPECT_RATIO;
        }
        else if (KP_PADDING_SYMMETRIC == inf_config->image_padding)
        {
            ncpu_img_config.image_format |= IMAGE_FORMAT_SYMMETRIC_PADDING;
        }
    }
    else // no pre-processing, bypass image
    {
        // no color space conversion, RGBA8888 only
        // no normalization
        // no scaling
        // no cropping

        ncpu_img_config.image_format |= IMAGE_FORMAT_BYPASS_PRE;
        int size = inf_config->image_buf_size;
        // copy image directly to model input addr
        // FIXME
        kdrv_gdma_memcpy((uint32_t)DDR_MEM_BASE, (uint32_t)ncpu_img_config.image_mem_addr, size, NULL, NULL);
    }

    dbg_print("ncpu image_format = 0x%x\n", ncpu_img_config.image_format);

    kmdw_model_config_img(&ncpu_img_config, p_crop_box, (struct kdp_pad_value_s *)inf_config->pad_value, inf_config->user_define_data);

    if (inf_config->enable_parallel)
    {
        kmdw_model_config_result(g_result_event, 0x1 << g_inf_index);

        g_result_ctx[g_inf_index].inf_result_buf = inf_config->inf_result_buf;
        g_result_ctx[g_inf_index].inf_result_buf_size = inf_config->inf_result_buf_size;
        g_result_ctx[g_inf_index].ncpu_result_buf = inf_config->ncpu_result_buf;
        g_result_ctx[g_inf_index].result_callback_func = inf_config->result_callback;

        g_num_parallel_inf++;

        g_inf_index++;
        if (g_inf_index >= MAX_OUTPUT_CONTEXT_NUM)
            g_inf_index = 0;
    }
    else
        kmdw_model_config_result(0, 0);

    dbg_print("ncpu_result_buf = 0x%x\n", inf_config->ncpu_result_buf);
    dbg_print("model_id = %d\n", inf_config->model_id);

    int status = kmdw_model_run("", inf_config->ncpu_result_buf, inf_config->model_id, true);

    if (status == IMAGE_STATE_TIMEOUT)
        return KP_FW_INFERENCE_TIMEOUT_103;
    else if (status != 0)
        return KP_FW_INFERENCE_ERROR_101;
    else
        return KP_SUCCESS;
}

void *kmdw_inference_app_result_get_free_buffer(int *buf_size)
{
    buffer_object_t fifoq_obj;

    // get output free buffer object for inference result
    // blocking wait, return only while a free result buffer is available
    dual_fifo2_get_free_buffer(_result_fifoq, &fifoq_obj, osWaitForever, false);

    if (buf_size)
        *buf_size = fifoq_obj.length;

    return (void *)fifoq_obj.buffer_addr;
}

osStatus_t kmdw_inference_app_result_enqueue(void *result_buf, int buf_size, bool preempt)
{
    buffer_object_t fifoq_output;
    fifoq_output.buffer_addr = (uint32_t)result_buf;
    fifoq_output.length = buf_size; // FIXME

    dbg_print("return result buffer 0x%x\n", fifoq_output.buffer_addr);
    return dual_fifo2_enqueue_data(_result_fifoq, fifoq_output, osWaitForever, preempt);
}

void kmdw_inference_app_send_status_code(int job_id, int error_code)
{
    // we need a result buffer
    int result_stamp_buf_size;
    kp_inference_header_stamp_t *result_stamp = (kp_inference_header_stamp_t *)kmdw_inference_app_result_get_free_buffer(&result_stamp_buf_size);

    result_stamp->magic_type = KDP2_MAGIC_TYPE_INFERENCE;
    result_stamp->total_size = sizeof(kp_inference_header_stamp_t);
    result_stamp->job_id = job_id;
    result_stamp->status_code = error_code;

    kmdw_inference_app_result_enqueue((void *)result_stamp, result_stamp_buf_size, false);
}

uint32_t kmdw_inference_app_get_model_raw_output_size(uint32_t model_id)
{
    // Warning: this raw output size is only for "data"

    // FIXME: should make this faster

    uint32_t *all_model_info = kmdw_model_get_all_model_info(true, false /*trust ddr data*/);
    if (!all_model_info)
        return 0;

    int num_models = all_model_info[0];
    for (int j = 0; j < num_models; j++)
    {
        struct kdp_model_s *model_info = kmdw_model_get_model_info(j);
        if (model_info->model_type == model_id)
        {
            return model_info->output_mem_len;
        }
    }

    return 0;
}

uint32_t kmdw_inference_get_raw_output_info_size(void)
{
    return kdp2_get_raw_output_info_size();
}

int kmdw_inference_app_init(kmdw_inference_app_callback_t app_entry, uint32_t image_count, uint32_t result_count)
{
    kmdw_printf("\n");
    kmdw_printf("starting KDP2 middleware ...\n");
    kmdw_printf("creating image queue with size %d\n", image_count);
    kmdw_printf("creating result queue with size %d\n", result_count);

    _image_fifioq = dual_fifo2_create(image_count);
    if ((uint32_t)_image_fifioq < DUAL_FIFO_VALID_ADDR)
    {
        kmdw_printf("image queue creating failed !!\n");
        return -1;
    }

    _result_fifoq = dual_fifo2_create(result_count);
    if ((uint32_t)_result_fifoq < DUAL_FIFO_VALID_ADDR)
    {
        kmdw_printf("result queue creating failed !!\n");
        return -1;
    }

    _app_entry_func = app_entry;

    g_result_event = osEventFlagsNew(0);

    return 0;
}

osStatus_t kmdw_inference_client_image_enqueue(uint32_t buf_addr, int buf_size, uint32_t timeout, bool preempt)
{
    buffer_object_t bobj;
    bobj.buffer_addr = buf_addr;
    bobj.length = buf_size;
    return dual_fifo2_enqueue_data(_image_fifioq, bobj, timeout, preempt);
}

osStatus_t kmdw_inference_client_result_request(uint32_t *buf_addr, int *buf_size, uint32_t timeout)
{
    buffer_object_t bobj;
    osStatus_t sts = dual_fifo2_dequeue_data(_result_fifoq, &bobj, timeout);
    *buf_addr = bobj.buffer_addr;
    *buf_size = bobj.length;
    return sts;
}

osStatus_t kmdw_inference_client_image_put_free_buffer(uint32_t buf_addr, int buf_size, uint32_t timeout)
{
    buffer_object_t bobj;
    bobj.buffer_addr = buf_addr;
    bobj.length = buf_size;
    return dual_fifo2_put_free_buffer(_image_fifioq, bobj, timeout);
}

osStatus_t kmdw_inference_client_image_get_free_buffer(uint32_t *buf_addr, int *buf_size, uint32_t timeout, bool force_grab)
{
    buffer_object_t bobj;
    osStatus_t sts = dual_fifo2_get_free_buffer(_image_fifioq, &bobj, timeout, force_grab);
    *buf_addr = bobj.buffer_addr;
    *buf_size = bobj.length;
    return sts;
}

osStatus_t kmdw_inference_client_result_put_free_buffer(uint32_t buf_addr, int buf_size, uint32_t timeout)
{
    buffer_object_t bobj;
    bobj.buffer_addr = buf_addr;
    bobj.length = buf_size;
    return dual_fifo2_put_free_buffer(_result_fifoq, bobj, timeout);
}

void kmdw_inference_client_clean_queues()
{
    buffer_object_t bobj;

    while (1)
    {
        if (dual_fifo2_dequeue_data(_image_fifioq, &bobj, 0) == osOK)
            dual_fifo2_put_free_buffer(_image_fifioq, bobj, 0);
        else
            break;
    }

    while (1)
    {
        if (dual_fifo2_dequeue_data(_result_fifoq, &bobj, 0) == osOK)
            dual_fifo2_put_free_buffer(_result_fifoq, bobj, 0);
        else
            break;
    }
}

