/*
 * Kneron DDR memory management utility
 *
 * Copyright (C) 2021 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include "cmsis_os2.h"
#include "base.h"
#include "kmdw_memory.h"
#include "kmdw_console.h"
#include "kmdw_ddr_mem_mgr.h"

#define FW_INFO_DDR_MASK         (1<<FW_INFO_DDR)
#define CAM_IMG_DDR_MASK         (1<<CAM_IMG_DDR)
#define FID_DDR_MASK             (1<<FID_DDR)
#define DB_DDR_MASK              (1<<DB_DDR)
#define SI_DDR_MASK              (1<<SI_DDR)
#define ISI_DDR_MASK             (1<<ISI_DDR)
#define FD_AGE_GENDER_DDR_MASK   (1<<FD_AGE_GENDER_DDR)
#define DME_DDR_MASK             (1<<DME_DDR)
#define JPEG_ENC_DDR_MASK        (1<<JPEG_ENC_DDR)
#define JPEG_DEC_DDR_MASK        (1<<JPEG_DEC_DDR)

typedef struct ddr_blk_link_s {
    struct ddr_blk_link_s *next;
    uint32_t addr;
}ddr_blk_link_t;

typedef struct inclusive_map_s {
    ddr_client_id_t id;

/*  the bit represent the clients mem blk can be inclusive with this client
    for example: JPEG_ENC_MASK|JPEG_DEC_MASK means Jpeg enc and jpeg dec DDR blocks are inclusive,
    this may support the application which runs jpeg enc and dec simutaneously  */
    uint32_t incl_mask_map;
}inclusive_map_t;

static bool bInitialized = FALSE;
static uint32_t ddr_clients_sts_mask;
static ddr_blk_link_t* ddr_blk_link_tbl[TOTAL_DDR_CLIENTS];

static inclusive_map_t  ddr_mem_incl_map_tbl[TOTAL_DDR_CLIENTS] = {
    {FW_INFO_DDR,       0xffff},  //compatible with all apps
    {MODEL_DDR,         0xffff},
    {CAM_IMG_DDR,       DME_DDR_MASK | FID_DDR_MASK | DB_DDR_MASK},
    {FID_DDR,           CAM_IMG_DDR_MASK | DB_DDR_MASK},
    {DB_DDR,            FID_DDR_MASK | CAM_IMG_DDR_MASK | ISI_DDR_MASK},
    {SI_DDR,            0},
    {ISI_DDR,           DB_DDR_MASK},
    {FD_AGE_GENDER_DDR, 0},
    {DME_DDR,           CAM_IMG_DDR_MASK},
    {JPEG_ENC_DDR,      0},
    {JPEG_DEC_DDR,      0},
    {HICO_DDR,          0},
};

/**********************************************************
                            Local functions
**********************************************************/
static void kmdw_ddr_mem_mgr_init(void)
{
    for (int i = 0; i < TOTAL_DDR_CLIENTS; i++)
    {
        ddr_blk_link_tbl[i] = NULL;
    }
    ddr_clients_sts_mask = 0;
}

static ddr_blk_link_t * reverse(ddr_blk_link_t* head)
{
    ddr_blk_link_t* current = head;
    ddr_blk_link_t *prev = NULL, *next = NULL;

    while (current != NULL) {
        next = current->next;
        current->next = prev;
        prev = current;
        current = next;
    }
    return prev;
}

/**********************************************************
                            Global functions
**********************************************************/
void kmdw_ddr_clean_client(ddr_client_id_t id)
{
    ddr_blk_link_t *head, *tail, *p_next;
    head = ddr_blk_link_tbl[id];
    tail = reverse(head);
    do {
        kmdw_ddr_release(tail->addr);
        p_next = tail->next;
        free(tail);
        tail = p_next;
    } while (tail != NULL);
    ddr_blk_link_tbl[id] = NULL;
    ddr_clients_sts_mask &= ~(1 << id);
}

uint32_t kmdw_ddr_malloc(ddr_client_id_t id, uint32_t numbyte)
{
    uint32_t mask;
    uint32_t inclusive_val;
    int i;
    ddr_blk_link_t *head;

    if (bInitialized == FALSE) {
        kmdw_ddr_mem_mgr_init();
        bInitialized = TRUE;
    }
    mask = 1 << id;
    if ((!(mask & ddr_clients_sts_mask)) && (ddr_blk_link_tbl[id] == NULL)) { /* not initialized */
        /* create link list */
        ddr_blk_link_tbl[id] = (ddr_blk_link_t *)malloc(sizeof(ddr_blk_link_t));
        memset(ddr_blk_link_tbl[id], 0, sizeof(ddr_blk_link_t));
        inclusive_val = ddr_mem_incl_map_tbl[id].incl_mask_map;

        /* free all DDR blocks owned by other exclusive cases */
        for (i = 0; i < TOTAL_DDR_CLIENTS; i++)
        {
            if (i == id)
                continue;
            uint32_t tmp = inclusive_val & (1 << i);
            if (tmp != 0)
                /* inclusive,  no need to free*/
                continue;

            /* exclusive each other, need to free */
            if (ddr_blk_link_tbl[i] == NULL)
                continue;

            kmdw_ddr_clean_client((ddr_client_id_t)i);
        }

        ddr_blk_link_tbl[id]->addr = kmdw_ddr_reserve(numbyte);
        if (ddr_blk_link_tbl[id]->addr == 0) {
            critical_msg("ddr_mem_mgr: failed to malloc DDR %d bytes\n", numbyte);
            return 0;
        }
        ddr_blk_link_tbl[id]->next = NULL;

        ddr_clients_sts_mask |= 1 << id;
        return ddr_blk_link_tbl[id]->addr;

    }
    else { // already be initialized
        head = ddr_blk_link_tbl[id];

        if (head == NULL) {
            critical_msg("Critical: NULL DDR Node, alloc failed\n");
            return 0;
        }

        while (head->next != NULL)
            head = head->next;

        head->next = (ddr_blk_link_t *)malloc(sizeof(ddr_blk_link_t));
        memset(head->next, 0, sizeof(ddr_blk_link_t));
        head->next->addr = kmdw_ddr_reserve(numbyte);
        if (head->next->addr == 0) {
            critical_msg("ddr_mem_mgr: failed to malloc DDR %d bytes\n", numbyte);
            return 0;
        }
        head->next->next = NULL;
        return head->next->addr;
    }
}

void kmdw_ddr_free(ddr_client_id_t id, uint32_t mem_addr)
{
    ddr_blk_link_t *temp, *prev, *head_ref;

    temp = ddr_blk_link_tbl[id];
    if (temp == NULL) {
        critical_msg("Error: this client has no memory to free\n");
        return;
    }

    if (temp->addr == mem_addr)
    {
        head_ref = temp->next;
        kmdw_ddr_release(mem_addr);
        free(temp);
        ddr_blk_link_tbl[id] = head_ref;

        if (head_ref == NULL) {
            ddr_clients_sts_mask &= ~(1 << id);
        }
        return;
    }

    while (temp != NULL && temp->addr != mem_addr)
    {
        prev = temp;
        temp = temp->next;
    }

    if (temp == NULL) {
        critical_msg("Error: cannot find this DDR memory\n");
        return;
    }

    prev->next = temp->next;
    kmdw_ddr_release(mem_addr);
    free(temp);
}

bool kmdw_ddr_is_mem_available(ddr_client_id_t id)
{
    if (ddr_blk_link_tbl[id] == NULL)
        return FALSE;
    else
        return TRUE;
}
