/*
 * Kneron Model API Manager
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "kdrv_ncpu.h"
#include "kmdw_ipc.h"
#include "kmdw_model.h"
#include "kmdw_console.h"   /*for dbg_msg */
#include "kmdw_memxfer.h"   /*for flash access */
#include "kmdw_memory.h"
#include "kmdw_ddr_mem_mgr.h"
#include "scu_ipc.h"        /*for NCPU triggering */
#include "kdp_crc.h"
#if defined(FLASH_TYPE) && (FLASH_TYPE == FLASH_TYPE_WINBOND_NAND)
#include "kdev_flash_nand.h"
#elif defined(FLASH_TYPE) && (FLASH_TYPE == FLASH_TYPE_GIGADEVICE_NAND)
#include "kdev_flash_nand.h"
#else
#include "kdev_flash.h"
#endif
#include "kdrv_clock.h"         /* for kdrv_delay_us() */
#include "project.h"

#define DEBUG 0
#define ENABLE_CHKSUM_ON_MODEL 0

#define FLAG_KMDW_MODEL_ABORT      BIT(29)  // Event flag to notify abort
#define FLAG_KMDW_MODEL_FROM_NCPU  BIT(30)  // Event flag to know NCPU is done
#define FLAG_KMDW_MODEL_FROM_NPU   BIT(28)  // Event flag to know NPU is done


#define KDP_FLASH_FW_INFO_SIZE      0x1000

extern const struct s_kdp_memxfer kdp_memxfer_module;

typedef struct {
    uint32_t n_model_source;           // 0: not set, 1: from flash, 2: from ddr
    uint32_t n_model_count;            // model count
    struct kdp_model_s p_model_info[KMDW_MODEL_MAX_MODEL_COUNT];  // save model info generated by compiler
    uint8_t pn_is_model_loaded_table[KMDW_MODEL_MAX_MODEL_COUNT]; // flag table to indicate if model is loaded
    uint32_t n_ddr_addr_model_end;     // DDR address of model end = user data start
    int32_t n_model_slot_index;        // scpu_to_ncpu->model_slot_index

    uint32_t    parallel_output_addr[KMDW_MODEL_MAX_MODEL_COUNT];
    uint32_t    ddr_addr_end;           // End DDR address including parallel output buffers
} kmdw_model_data_t;

static kmdw_model_data_t s_model_data = {0};

typedef struct {
    int32_t raw_img_idx;
    osEventFlagsId_t evt_caller;        // event to know/control ncpu
    uint32_t        caller_e;
    osEventFlagsId_t evt_result;        // event to know/control npu
    uint32_t        result_e;
} kmdw_img_data_t;

// buf for uploaded model from host
static char s_fw_info_buf[KDP_FLASH_FW_INFO_SIZE];

// ptr to the buf for uploaded fw info from host
static kmdw_model_fw_info_t *s_fw_info_buf_p = (kmdw_model_fw_info_t *)s_fw_info_buf;
// ptr to the buf for uploaded fw info from flash
static kmdw_model_fw_info_t *s_fw_info_buf_for_flash_p = NULL;

static kmdw_img_data_t s_img_data[IPC_IMAGE_ACTIVE_MAX] = {0};
static int32_t s_current_ipc_idx = 0;
static int32_t s_next_ipc_idx = 0;
static int32_t model_paralleled[KMDW_MODEL_MAX_MODEL_COUNT] = {0};


/*  jpeg related definitions   */
typedef struct model_jpeg_enc_ctx_s {
    scpu_to_ncpu_t*         p_ipc_out;
    ncpu_to_scpu_result_t*  p_ipc_in;
    osEventFlagsId_t        jpeg_enc_evt_id;
    uint32_t                jpeg_enc_evt_flag;
    jpeg_encode_result_t    npu_rslt;
} jpeg_enc_kmdw_ctx_t;

static jpeg_enc_kmdw_ctx_t jpeg_enc_kmdw_ctx = {0};

typedef struct model_jpeg_dec_ctx_s {
    scpu_to_ncpu_t*         p_ipc_out;
    ncpu_to_scpu_result_t*  p_ipc_in;
    osEventFlagsId_t         jpeg_dec_evt_id;
    uint32_t                        jpeg_dec_evt_flag;
    jpeg_decode_result_t    npu_rslt;
} jpeg_dec_kmdw_ctx_t;

static jpeg_dec_kmdw_ctx_t jpeg_dec_kmdw_ctx = {0};

/* ############################
 * ##    Static Functions    ##
 * ############################ */


/**
 * @brief load fw info from flash
 * @return 0: OK, -1: fail
 * @note NULL means failed; non-zero ptr means OK
 */
static kmdw_model_fw_info_t* _load_flash_model_info(void)
{
    //load model from flash once, and reuse loaded data

    if ((NULL == s_fw_info_buf_for_flash_p ) || !kmdw_ddr_is_mem_available(FW_INFO_DDR)) {
        s_fw_info_buf_for_flash_p = (kmdw_model_fw_info_t*)kmdw_ddr_malloc(FW_INFO_DDR, KDP_FLASH_FW_INFO_SIZE);

        if (NULL == s_fw_info_buf_for_flash_p)
            critical_msg("[ERR] insufficent memory for reading fw_info from flash\n");
        else {
            kdp_memxfer_module.flash_to_ddr((uint32_t)s_fw_info_buf_for_flash_p, FLASH_MODEL_FW_INFO_ADDR, KDP_FLASH_FW_INFO_SIZE);
        }
    }

    return s_fw_info_buf_for_flash_p;
}

/**
 * @brief set s_model_data.parallel_output_addr and ddr_addr_end
 * @return 0: OK, -1: fail
 */
static int32_t _set_parallel_output_addr_and_ddr_end(void)
{
    int i;
    uint32_t addr, len, len_max = 0;

    if (s_model_data.n_model_count >= KMDW_MODEL_MAX_MODEL_COUNT) {
        err_msg("[ERR] model count %d exceeds max %d\n",
                s_model_data.n_model_count, KMDW_MODEL_MAX_MODEL_COUNT);
        return 0;
    }

    addr = ALIGN64(s_model_data.n_ddr_addr_model_end);
    len = 0;
    for (i = 0; i < s_model_data.n_model_count; i++) {
        addr = ALIGN64(addr + len);
        s_model_data.parallel_output_addr[i] = addr;

        len = s_model_data.p_model_info[i].output_mem_len;
        if (len > len_max )
            len_max = len;
        dbg_msg("[DBG] model[%d]: parallel buffer: 0x%x (len 0x%x)\n", i, addr, len);

        // Reset parallel handling
        model_paralleled[i] = 0;
    }
    dbg_msg("[DBG] max parallel buffer: 0x%x (len 0x%x)\n", addr, len_max);
    addr = ALIGN64(addr + len_max);

    if (addr >= DDR_HEAP_END) {
        err_msg("[ERR] model: parallel buffer: 0x%x over (>=) boundary 0x%x\n", addr, DDR_HEAP_END);
        return -1;
    } else {
        dbg_msg("[DBG] model & info DDR end: 0x%x (0x%x)\n", addr, s_model_data.n_ddr_addr_model_end);
    }

    s_model_data.ddr_addr_end = addr;
    return 0;
}

/**
 * @brief reset s_model_data
 */
static void _reset_model_data(void)
{
    s_model_data.n_model_count = 0;
    s_model_data.n_model_source = 0;
    memset( s_model_data.p_model_info, 0, sizeof(s_model_data.p_model_info));
    memset( s_model_data.pn_is_model_loaded_table, 0, sizeof(s_model_data.pn_is_model_loaded_table));

    return;
}

/**
 * @brief convert modeltype to modelInfo array index
 * @param model_type_p: model type (defined in model_type.h)
 * @return modelInfo model index (starts from 0)
 *         -1 means not such modeltype in flash
 */
static int8_t _get_model_info_array_index_by_model_type(uint32_t model_type_p)
{
    int i;
    for(i=0 ; i < s_model_data.n_model_count; i++) {
        if(s_model_data.p_model_info[i].model_type == model_type_p)
            return i;
    }

    return -1;
}

/**
 * @brief get fw info extension data from fw_info ptr
 * @param[in] fw_info_p the ptr to fw_info
 * @return the ptr to fw_info_ext
 */
static kmdw_model_fw_info_ext_t*
_get_fw_info_ext_by_fw_info(kmdw_model_fw_info_t* fw_info_p)
{
    if(NULL == fw_info_p)
        return NULL;
    else {
        kmdw_model_fw_info_ext_t* ret = NULL;
        uint32_t count;
        uint32_t offset;

        count = fw_info_p->model_count;
        offset = sizeof(struct kdp_model_s) * count;
        ret = (kmdw_model_fw_info_ext_t *)((uint32_t)fw_info_p->models + offset);
        return ret;
    }
}

/**
 * @brief load model information generated by compiler
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 * @param [in] is_reload        : is force reload
 * @return model count
 *         0 means no model is loaded in this call
 */
static int32_t _load_model_info(bool from_ddr, bool reload)
{
#if ENABLE_CHKSUM_ON_MODEL
    u32 sum32;
    u8 *addr;
#endif

    if (s_model_data.n_model_count && !reload) {
        return s_model_data.n_model_count;
    }

    if (reload) {
        _reset_model_data();
    }

    kmdw_model_fw_info_t *model_info_p = NULL;
    kmdw_model_fw_info_ext_t *model_info2_p = NULL;

    // load model Info
    if (from_ddr) {
        model_info_p = s_fw_info_buf_p;
        model_info2_p = _get_fw_info_ext_by_fw_info(model_info_p);

        if((NULL == model_info_p) || (NULL == model_info2_p) ) {
            s_model_data.n_model_count = 0;
            return 0;
        }

        // get model count
        s_model_data.n_model_count = model_info_p->model_count;
        dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);

        if(0 == s_model_data.n_model_count) {
            info_msg("[info] model is not in DDR!!\n");
            return 0;
        } else if (s_model_data.n_model_count > KMDW_MODEL_MAX_MODEL_COUNT) {
            info_msg("[ERR] model count is over MAX limit=%d!!\n", KMDW_MODEL_MAX_MODEL_COUNT);
            s_model_data.n_model_count = 0;
            return 0;
        } else {
            dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);
        }

        // get model info
        memcpy(s_model_data.p_model_info, (const void*)model_info_p->models,
               sizeof(struct kdp_model_s)*s_model_data.n_model_count);

        // get ddr model end addr
        s_model_data.n_ddr_addr_model_end = model_info2_p->model_dram_addr_end;
        if (s_model_data.n_ddr_addr_model_end >= DDR_HEAP_END) {
            err_msg("[ERR] modelInfo: DDR end address: 0x%x over (>=) boundary 0x%x\n", s_model_data.n_ddr_addr_model_end, DDR_HEAP_END);
            return 0;
        } else {
            dbg_msg("[DBG] modelInfo: DDR end address: 0x%x\n", s_model_data.n_ddr_addr_model_end);
        }

        // set model source
        s_model_data.n_model_source = 2; // from ddr
    } else { // models are stored in flash

        model_info_p = _load_flash_model_info();  // this function updates data on s_fw_info_buf_for_flash_p
        model_info2_p = _get_fw_info_ext_by_fw_info(model_info_p);

        if((NULL == model_info_p) || (NULL == model_info2_p) ) {
            s_model_data.n_model_count = 0;
            return 0;
        }

        // get model count
        s_model_data.n_model_count = model_info_p->model_count;
        dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);

        if (s_model_data.n_model_count == 0xFFFFFFFF) {
            err_msg("[info] model is not in flash!!\n");
            s_model_data.n_model_count = 0;
            return 0;
        } else if (s_model_data.n_model_count > KMDW_MODEL_MAX_MODEL_COUNT) {
            info_msg("[ERR] model count is over MAX limit=%d!!\n", KMDW_MODEL_MAX_MODEL_COUNT);
            s_model_data.n_model_count = 0;
            return 0;
        } else {
            dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);
        }

        // get model info
        //FIXME, why need to clone to s_fw_info_buf_p
        memcpy(s_fw_info_buf_p, (void *)model_info_p, KDP_FLASH_FW_INFO_SIZE);

        memcpy(s_model_data.p_model_info, model_info_p->models, sizeof(struct kdp_model_s)*s_model_data.n_model_count);

#if ENABLE_CHKSUM_ON_MODEL
        addr = (u8 *)s_model_data.p_model_info;
        sum32 = kdp_gen_sum32(addr, sizeof(struct kdp_model_s)*s_model_data.n_model_count);
        if(MODEL_INFO_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model_info wrong sum 0x%08x/correct 0x%x!!!~~~\n",
                         sum32, MODEL_INFO_SUM32_VALUE);
        }
#endif
        // get ddr model end addr
        s_model_data.n_ddr_addr_model_end = model_info2_p->model_dram_addr_end;

        if (s_model_data.n_ddr_addr_model_end >= DDR_HEAP_END) {
            err_msg("[ERR] modelInfo: DDR end address: 0x%x over (>=) boundary 0x%x\n", s_model_data.n_ddr_addr_model_end, DDR_HEAP_END);
            return 0;
        } else {
            dbg_msg("[DBG] modelInfo: DDR end address: 0x%x\n", s_model_data.n_ddr_addr_model_end);
        }

        // set model source
        s_model_data.n_model_source = 1; // from flash
    }

    int ret = _set_parallel_output_addr_and_ddr_end();
    if(-1 == ret)
        return 0;

    return s_model_data.n_model_count;
}

/**
 * @brief load specific model by model info index (the order in flash)
 * @param model_index_p: model info index
 * @return 0: model not ready, 1: model is loaded
 */
static int32_t _load_model(uint8_t model_index_p/*starts from 0*/)
{
    uint32_t ddr_addr_models_head; //start point = the 1st model's cmd.bin
    uint32_t ddr_addr_offset;
    uint32_t flash_addr;
    uint32_t len_to_load;
#if ENABLE_CHKSUM_ON_MODEL
    u32 sum32;
    u8 *addr;
#endif
    struct kdp_model_s *p_model;

    if(s_model_data.n_model_count == 0)
        return 0; // model info is not ready

    if(s_model_data.pn_is_model_loaded_table[model_index_p] == 1 )
        return 1; //model has been loaded
    else
        s_model_data.pn_is_model_loaded_table[model_index_p] = 1;

    //load model with (index=model_index_p) from flash to DDR
    ddr_addr_models_head = s_model_data.p_model_info[0].cmd_mem_addr; //start point = the 1st model's cmd.bin

    //load cmd + weight + setup together
    p_model = &(s_model_data.p_model_info[model_index_p]);
    ddr_addr_offset = p_model->cmd_mem_addr - ddr_addr_models_head;

    flash_addr = FLASH_MDDEL_ALL_ADDR + ddr_addr_offset;

    len_to_load = ALIGN16(p_model->cmd_mem_len) +
                  ALIGN16(p_model->weight_mem_len) +
                  ALIGN16(p_model->setup_mem_len);

    //model from flash to ddr
    kdp_memxfer_module.flash_to_ddr(p_model->cmd_mem_addr, flash_addr, len_to_load);

#if ENABLE_CHKSUM_ON_MODEL
    dbg_msg("[INFO] %s: load model[%d] at ddr[0x%x], sum32:0x%x\n",
            __FUNCTION__, model_index_p, p_model->cmd_mem_addr, sum32);

    /*Calculate & compare check value */
    addr = (u8 *)p_model->cmd_mem_addr;
    sum32 = kdp_gen_sum32(addr, len_to_load);
#ifdef FLASH_MODEL_HAS_CRC_CHECK  // no check field is defined yet, this section won't work
    if (p_model->crc16 != crc16) {
        critical_msg("[ERR] model[%d] wrong crc 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                     model_index_p, crc16, p_model->crc16, addr, len_to_load);
    }
#else  // this is only okay for default flash image
    // FIXME, should be removed because default example is not FDR anymore
    switch(model_index_p) {
    case 0: //FD
        if(MODEL0_FD_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL0_FD_SUM32_VALUE, addr, len_to_load);
        }
        break;
    case 1: //LM
        if(MODEL1_LM_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL1_LM_SUM32_VALUE, addr, len_to_load);
        }
        break;
    case 3: //FR
        if(MODEL3_FR_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL3_FR_SUM32_VALUE, addr, len_to_load);
        }
        break;
    }
#endif
#endif
    return 1;
}


/**
 * @brief specify model information, load model info, load model
 * @param [in] model_type_p: model unique ID defined by Kneron
 * @param [in] model_from_ddr: is model from ddr or host command
 * @return model_slot_index(requested by NCPU/NPU)
 *         -1 : model not found
 */
static int32_t _config_model(uint32_t model_type, bool model_from_ddr)
{
    int model_info_idx; //limitation (hard coded in flash)
    int model_idx;

    //check if model info is loaded
    if( 0 == _load_model_info(model_from_ddr, false/*reload*/)) {
        return -1;
    }

    if( model_from_ddr == 0 ) {
        //FIXME, should remove application related code
        /* Special model not in DDR but in ncpu */
        if (model_type == KNERON_2D_LIVENESS_V3_FACEBAGNET_224_224_3) {
            model_idx = 3;
            model_info_idx = 4;

            goto model_common;
        }

        model_info_idx = _get_model_info_array_index_by_model_type(model_type);
        if(model_info_idx == -1) {
            err_msg("[ERR] model_type[%d] is not found in flash\n", model_type);
            return -1;
        }
        _load_model(model_info_idx);

        // FIXME: need to remove the following hard code
        model_idx = model_info_idx;

    } else {
        model_info_idx = _get_model_info_array_index_by_model_type(model_type);
        if(model_info_idx == -1) {
            err_msg("[ERR] model_type[%d] is not found in DDR\n", model_type);
            return -1;
        }
        model_idx = model_info_idx;
    }

model_common:
    s_model_data.n_model_slot_index = model_idx;

    kmdw_ipc_set_model(s_model_data.p_model_info, model_info_idx, model_idx);

    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(s_img_data[s_current_ipc_idx].raw_img_idx);

    if (raw_img->format & IMAGE_FORMAT_PARALLEL_PROC) {
        int mirrored_idx = IPC_MODEL_MAX - model_idx - 1;      // mirror for parallel

        if (!model_paralleled[model_idx]) {

            model_paralleled[model_idx] = 1;

            /* Duplicate model_info and use parallel output buffer */
            struct scpu_to_ncpu_s* comm_out = kmdw_ipc_get_output();
            struct kdp_model_s *p_mirror_model_info = s_model_data.p_model_info;
            comm_out->models[mirrored_idx] = *(p_mirror_model_info + model_idx);
            comm_out->models_type[mirrored_idx] = (p_mirror_model_info + model_idx)->model_type;

            uint32_t len = comm_out->models[mirrored_idx].output_mem_len;
            comm_out->models[mirrored_idx].output_mem_addr = s_model_data.parallel_output_addr[model_idx];

            info_msg("Parallel buffer: model %d [%d %d], len %d, addr 0x%x\n",
                     model_type, model_idx, mirrored_idx, len,
                     comm_out->models[mirrored_idx].output_mem_addr);
        }

        if (s_current_ipc_idx != 0)
            model_idx = mirrored_idx;
    }

    struct scpu_to_ncpu_s* p_comm_out = kmdw_ipc_get_output();

    if(!kmdw_ddr_is_mem_available(MODEL_DDR)) {
        p_comm_out->output_mem_addr3 = 0;
        p_comm_out->output_mem_addr4 = 0;
    }

    if ((p_comm_out->output_mem_addr3 == 0) || !kmdw_ddr_is_mem_available(MODEL_DDR)) {
        uint32_t len = 0x5000;
        p_comm_out->output_mem_addr3 = kmdw_ddr_malloc(MODEL_DDR, len*sizeof(uint32_t));
        if(p_comm_out->output_mem_addr3 == 0) {
            critical_msg("kmdw_model: failed to malloc comm_out->output_mem_addr3\n");
            return -1;
        }
    }
    if ((p_comm_out->output_mem_addr4 == 0) || !kmdw_ddr_is_mem_available(MODEL_DDR)) {
        uint32_t len = 8 * (1 << 20);
        p_comm_out->output_mem_addr4 = kmdw_ddr_malloc(MODEL_DDR, len);
        if(p_comm_out->output_mem_addr4 == 0) {
            critical_msg("kmdw_model: failed to malloc comm_out->output_mem_addr4\n");
            return -1;
        }
        info_msg("CenterNet/Yolo5 (%d) buffer: 0x%x (len 0x%x)\n", model_type, p_comm_out->output_mem_addr4, len);
    }

    kmdw_ipc_set_model_active(model_idx);

    return model_idx;
}


/**
 * @brief run model according to config settings
 * @return status defined in NCPU
 * @note !!! must be called after kapp_config_model_image()
 */
static int32_t _run_model(void)
{
    int active_idx = s_current_ipc_idx;
    int raw_img_idx = s_img_data[active_idx].raw_img_idx;
    struct kdp_img_raw_s *p_raw_image = kmdw_model_get_raw_img(raw_img_idx);
    uint32_t flags, wait_evt;
    uint32_t is_abort = 0;

    // Start time for ncpu/npu round trip
    p_raw_image->tick_start = osKernelGetSysTimerCount();

    if (s_img_data[active_idx].evt_caller == NULL)
        s_img_data[active_idx].evt_caller = osEventFlagsNew(0);

    // set notify for job done
    if (s_img_data[active_idx].evt_result) {
        /* Result event already set. Let's do local event for parallel. */
        wait_evt = FLAG_KMDW_MODEL_FROM_NPU;
    } else {
        wait_evt = FLAG_KMDW_MODEL_FROM_NCPU;
    }

    dbg_msg("<Run-Model> wait %d[%d] evt %x\n", raw_img_idx, active_idx, wait_evt);

    //assign caller event before triggering ncpu/npu
    s_img_data[active_idx].caller_e = wait_evt;

    //trigger ncpu/npu
    kmdw_ipc_trigger_int(CMD_RUN_NPU);

    p_raw_image->tick_got_ncpu_ack = osKernelGetSysTimerCount();

    //check abort signal
    flags = osEventFlagsWait(s_img_data[active_idx].evt_caller,
                             FLAG_KMDW_MODEL_ABORT,
                             osFlagsWaitAll, 0);
    if( flags != osFlagsErrorResource ) {
        osEventFlagsClear(s_img_data[active_idx].evt_caller, FLAG_KMDW_MODEL_ABORT);
        is_abort = 1;
    }

#define MODEL_INF_TIMEOUT (2000) // 2 secs timeout OK ? FIXME

    //wait for finish of current task
    flags = osEventFlagsWait(s_img_data[active_idx].evt_caller,
                             wait_evt,
                             osFlagsNoClear, MODEL_INF_TIMEOUT);
    if(flags == osFlagsErrorTimeout) {
        err_msg("[%s] osEventFlagsWait flag 0x%08x timeout\n", __FUNCTION__, wait_evt);
        return IMAGE_STATE_TIMEOUT;
    } else if (flags != wait_evt)
        dbg_msg("[%s] 1+ events 0x%08x (%d[%d] expected)\n", __FUNCTION__, flags, wait_evt, active_idx);
    else
        dbg_msg("[DBG][%s] got: raw_img_idx[active_idx]=%d[%d]\n", __FUNCTION__, raw_img_idx, active_idx);

    osEventFlagsClear(s_img_data[active_idx].evt_caller, wait_evt);

    if( 1 == is_abort ) {
        dbg_msg("[DBG][%s] abort after n_model_slot_index = %d\n", __FUNCTION__, s_model_data.n_model_slot_index);
        return KMDW_MODEL_RUN_RC_ABORT; //abort
    }

    //FIXME, should return sts
    return 0;
}

#ifndef KDP2_FW
// sorry, due to extern variable
extern u16 APP_id;
#endif

static void _ipc_handler(struct kdp_img_raw_s *p_raw_image, int state)
{
    int ipc_idx;

    if (state == IMAGE_STATE_RECEIVING) {
        ipc_idx = p_raw_image->ref_idx;

        // End time for ncpu/npu round trip
        p_raw_image->tick_end = osKernelGetSysTimerCount();

        if (s_img_data[ipc_idx].evt_result) {
            ipc_dbg_msg("[done: post: P] ipc_idx: %d, result_e: %d (ram %x)\n", ipc_idx, s_img_data[ipc_idx].result_e, p_raw_image);
            osEventFlagsSet(s_img_data[ipc_idx].evt_result, s_img_data[ipc_idx].result_e);
        } else {
            ipc_dbg_msg("[done: post: S] ipc_idx: %d, caller_e: %x.\n", ipc_idx, s_img_data[ipc_idx].caller_e);
            osEventFlagsSet(s_img_data[ipc_idx].evt_caller, s_img_data[ipc_idx].caller_e);
        }
    } else if (state == IMAGE_STATE_ACTIVE) {
        ipc_idx = s_current_ipc_idx;
        ipc_dbg_msg("[done: npu: P] ipc_idx: %d, caller_e: %x\n", ipc_idx, s_img_data[ipc_idx].caller_e);
        osEventFlagsSet(s_img_data[ipc_idx].evt_caller, s_img_data[ipc_idx].caller_e);
    } else if( state == IMAGE_STATE_JPEG_ENC_DONE) {
        ncpu_to_scpu_result_t * p_in_ipc = jpeg_enc_kmdw_ctx.p_ipc_in;
        jpeg_encode_result_t  *pRes = (jpeg_encode_result_t  *)p_in_ipc->extRsltAddr;
        jpeg_enc_kmdw_ctx.npu_rslt = *pRes;
        if(jpeg_enc_kmdw_ctx.jpeg_enc_evt_id == 0) {
            ipc_err_msg("jpeg_enc_evt_id was not created\n");
            return;
        }
        osEventFlagsSet(jpeg_enc_kmdw_ctx.jpeg_enc_evt_id, jpeg_enc_kmdw_ctx.jpeg_enc_evt_flag);
    } else if( state == IMAGE_STATE_JPEG_DEC_DONE) {
        ncpu_to_scpu_result_t * p_in_ipc = jpeg_dec_kmdw_ctx.p_ipc_in;
        jpeg_decode_result_t  *pRes = (jpeg_decode_result_t  *)p_in_ipc->extRsltAddr;
        jpeg_dec_kmdw_ctx.npu_rslt = *pRes;
        if(jpeg_dec_kmdw_ctx.jpeg_dec_evt_id == 0) {
            ipc_err_msg("jpeg_dec_evt_id was not created\n");
            return;
        }
        osEventFlagsSet(jpeg_dec_kmdw_ctx.jpeg_dec_evt_id, jpeg_dec_kmdw_ctx.jpeg_dec_evt_flag);
    } else {
        ipc_err_msg("[IPC ERR] wrong state: %d\n", state);
    }
}





/**
 * @brief common code of getting fw_info
 * @param[in] from_ddr true:to get fw_info from loaded in ddr
 *                     flase:to get fw_info from flash (load model to temp ddr)
 * @param[in] trust_ddr_data if trust ddr data
 * @return address of loaded fw_info
 * @note DDR data might be dirty, so use trust_ddr_data if make sure ddr data is valid
 */
static kmdw_model_fw_info_t* _get_fw_info(bool from_ddr, bool trust_ddr_data)
{
    uint32_t model_cnt;
    kmdw_model_fw_info_t *fw_info_ptr = NULL;

    if (false == from_ddr){
        fw_info_ptr = _load_flash_model_info();
        if (NULL == fw_info_ptr) {
            err_msg("[ERR] insufficent memory for reading data from flash\n");
        }
    } else if ((true == from_ddr) && (true == trust_ddr_data)) {
        fw_info_ptr = s_fw_info_buf_p;
    } else {
        if (0 < s_model_data.n_model_count) {
            if (1 == s_model_data.n_model_source) { // s_mode_data is loaded from flash
                fw_info_ptr = _load_flash_model_info();
                if (NULL == fw_info_ptr) {
                    err_msg("[ERR] insufficent memory for reading data from flash\n");
                }
            } else if (2 == s_model_data.n_model_source) {  // s_mode_data is loaded from ddr
                // The flow has parsed DDR data to s_model_data, so DDR data is valid
                fw_info_ptr = s_fw_info_buf_p;
            } else { // s_model_data is not loaded
                fw_info_ptr = NULL;
            }
        } else {
            // s_model_data is empty means DDR might be old data before reset
            fw_info_ptr = NULL;
        }
    }

    if(fw_info_ptr) {
        model_cnt = fw_info_ptr->model_count;

        if( 0 == model_cnt || model_cnt > KMDW_MODEL_MAX_MODEL_COUNT) {
            if (from_ddr)
                dbg_msg("Model is not in DDR\n");
            else
                dbg_msg("Model is not in Flash\n");

            return 0;
        }
    }

    return fw_info_ptr;
}

/* ############################
 * ##    Public Functions    ##
 * ############################ */

void kmdw_model_init(void)
{
    kmdw_ipc_initialize(_ipc_handler);
    memset( (void *)&jpeg_enc_kmdw_ctx, 0, sizeof(jpeg_enc_kmdw_ctx));

#ifndef KDP2_FW
// sorry, due to extern variables
    jpeg_4M_reserved_blk = 0;
    jpeg_2M_reserved_blk = 0;
#endif

    s_fw_info_buf_p->model_count = 0;
}

int32_t kmdw_model_load_model(int8_t model_info_index_p)
{
    int32_t ret = 0;

    if(1 != s_model_data.n_model_source ||  // check if s_model_data is not according to flash
       0 == s_model_data.n_model_count) {
        if(0 ==  _load_model_info(false/*from ddr*/, true/*reload*/))
            return 0; //error, no model is loaded
    }

    // load all models
    if (KMDW_MODEL_ALL_MODELS == model_info_index_p) {
        uint8_t i;
        for (i = 0 ; i < s_model_data.n_model_count ; i++) {
            ret = _load_model(i);
            if( 0 == ret) {
                err_msg("[ERR] %s : failed to load model array index:%d\n", __FUNCTION__, i);
                return 0;
            }
        }
        return s_model_data.n_model_count;
    } else { // load specific model
        ret = _load_model(model_info_index_p);
        return ret;
    }
}

int32_t kmdw_model_reload_model_info(bool from_ddr)
{
    return _load_model_info(from_ddr, true/*reload*/);
}

int32_t kmdw_model_refresh_models(void)    // reload all the models from flash again
{
    uint8_t i;

    // forcedly update s_model_data which might be poluted by model upload from host
    if(0 ==  _load_model_info(false/*from ddr*/, true/*reload*/))
        return 0; //error, no model is loaded

    int ret;
    for (i = 0 ; i < s_model_data.n_model_count ; i++) {
        if (s_model_data.pn_is_model_loaded_table[i]) {  // if previously loaded
            s_model_data.pn_is_model_loaded_table[i] = 0;
            ret = _load_model(i);  // reload the model again
            if ( 0 == ret) {
                err_msg("[ERR] %s : failed to load model array index:%d\n", __FUNCTION__, i);
                return 0;
            }
        }
    }
    return s_model_data.n_model_count;
}

int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag)
{
    int active_idx = s_current_ipc_idx;

    s_img_data[active_idx].evt_result = result_evt;
    s_img_data[active_idx].result_e = result_evt_flag;

    return 0;
}

void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param)
{
    int act_img_idx = img_cfg->image_buf_active_index;
    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(act_img_idx);

    s_current_ipc_idx = s_next_ipc_idx;

    if (img_cfg->image_format & IMAGE_FORMAT_PARALLEL_PROC)
        s_next_ipc_idx = !s_next_ipc_idx;

    kmdw_ipc_set_image_active(act_img_idx);
    s_img_data[s_current_ipc_idx].raw_img_idx = act_img_idx;

    raw_img->state = IMAGE_STATE_ACTIVE;
    raw_img->seq_num = act_img_idx;

    raw_img->ref_idx = s_current_ipc_idx;
    raw_img->input_row = img_cfg->image_row;
    raw_img->input_col = img_cfg->image_col;
    raw_img->input_channel = img_cfg->image_ch;
    raw_img->format = img_cfg->image_format;
    raw_img->image_mem_addr = img_cfg->image_mem_addr;
    raw_img->image_mem_len = img_cfg->image_mem_len;

    if (crop_box == NULL) {
        raw_img->params_s.crop_top = 0;
        raw_img->params_s.crop_bottom = 0;
        raw_img->params_s.crop_left = 0;
        raw_img->params_s.crop_right = 0;
    } else {
        raw_img->params_s.crop_top = crop_box->top;
        raw_img->params_s.crop_bottom = crop_box->bottom;
        raw_img->params_s.crop_left = crop_box->left;
        raw_img->params_s.crop_right = crop_box->right;
    }

    if (pad_values == NULL) {
        raw_img->params_s.pad_top = 0;
        raw_img->params_s.pad_bottom = 0;
        raw_img->params_s.pad_left = 0;
        raw_img->params_s.pad_right = 0;
    } else {
        raw_img->params_s.pad_top = pad_values->pad_top;
        raw_img->params_s.pad_bottom = pad_values->pad_bottom;
        raw_img->params_s.pad_left = pad_values->pad_left;
        raw_img->params_s.pad_right = pad_values->pad_right;
    }

    if (ext_param == NULL) {
        memset(raw_img->params_s.params, 0, MAX_PARAMS_LEN*4);
    } else {
        memcpy(raw_img->params_s.params, ext_param, MAX_PARAMS_LEN*4);
    }
}

struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx)
{
    struct scpu_to_ncpu_s *comm_out = kmdw_ipc_get_output();
    return &(comm_out->raw_images[idx]);
}

int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool model_from_ddr)
{
    int model_idx = _config_model(model_type, model_from_ddr);
    if (model_idx < 0) {
        return KMDW_MODEL_RUN_RC_ABORT;
    }

    int img_idx = s_img_data[s_current_ipc_idx].raw_img_idx;
    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(img_idx);

//    raw_img->results[model_idx].result_mem_addr = (uint32_t)output;
    raw_img->result.result_mem_addr = (uint32_t)output;

    dbg_msg("[INFO] %s:\n", tag);
    dbg_msg(" model_idx = %d\n", model_idx);
    dbg_msg("       (row/col/ch) = %d/%d/%d\n", raw_img->input_row, raw_img->input_col, raw_img->input_channel);
    dbg_msg("             format = 0x%x\n", raw_img->format);
    dbg_msg("  crop(tp/bt/lf/rt) = %d/%d/%d/%d\n", raw_img->params_s.crop_top, raw_img->params_s.crop_bottom,
            raw_img->params_s.crop_left, raw_img->params_s.crop_right);
    dbg_msg("*ext_params(first 4)= %d/%d/%d/%d\n", raw_img->params_s.params[0], raw_img->params_s.params[1],
            raw_img->params_s.params[2], raw_img->params_s.params[3]);
    dbg_msg("            ref_idx = %d\n", raw_img->ref_idx);
    dbg_msg("         model type = %d\n", model_type);
    dbg_msg("        image  addr = 0x%x\n", raw_img->image_mem_addr);
    dbg_msg("        output addr = 0x%x\n", raw_img->result.result_mem_addr);

    return _run_model();
}

void kmdw_model_abort(void)
{
    int active_idx = s_current_ipc_idx;

    if( 0 == s_img_data[active_idx].evt_caller)
        return;

    osEventFlagsSet(s_img_data[active_idx].evt_caller, FLAG_KMDW_MODEL_ABORT);
}

struct kdp_model_s* kmdw_model_get_model_info(int model_idx_p)
{
    if (s_model_data.n_model_count == 0) {
        return NULL;
    } else if (model_idx_p >= s_model_data.n_model_count) {
        return NULL;
    } else {
        return &(s_model_data.p_model_info[model_idx_p]);
    }
}

void kmdw_model_get_run_time(int img_idx, kmdw_model_run_time_t *run_time/*out*/)
{
    uint32_t  freq_us;
    struct kdp_img_raw_s *p_raw_image;
    static uint32_t tick_prev_end = 0;

    if (run_time == NULL)
        return;


    p_raw_image = kmdw_model_get_raw_img(img_idx);
    freq_us = osKernelGetSysTimerFreq()/1000000;
    run_time->round_trip_time = (p_raw_image->tick_end - p_raw_image->tick_start)/(freq_us*1000.0);
    run_time->rslt_pop_interval = (p_raw_image->tick_end - tick_prev_end)/(freq_us*1000.0);
    run_time->scpu_send_ipc_ack = (p_raw_image->tick_got_ncpu_ack - p_raw_image->tick_start)/(freq_us*1000.0);
#ifdef PERF_HIGH_ACCURACY
    run_time->pre_proc_time = (p_raw_image->tick_end_pre - p_raw_image->tick_start_pre)*1.0/NCPU_CLOCK_CNT_PER_MS;
    run_time->npu_proc_time = (p_raw_image->tick_cnn_interrupt_rvd - p_raw_image->tick_start_npu)*1.0/NCPU_CLOCK_CNT_PER_MS;
    run_time->post_proc_time = (p_raw_image->tick_end_post - p_raw_image->tick_start_post)*1.0/NCPU_CLOCK_CNT_PER_MS;
    run_time->ncpu_req_ipc_ack = (p_raw_image->tick_ncpu_img_ack - p_raw_image->tick_ncpu_img_req)*1.0/NCPU_CLOCK_CNT_PER_MS;
    run_time->ncpu_rslt_ipc_ack = (p_raw_image->tick_rslt_got_scpu_ack - p_raw_image->tick_end_post)*1.0/NCPU_CLOCK_CNT_PER_MS;
    run_time->ncpu_img_req_to_rcv_new_img = (p_raw_image->tick_start_pre - p_raw_image->tick_last_img_req)*1.0/NCPU_CLOCK_CNT_PER_MS;
#else
    run_time->pre_proc_time = p_raw_image->tick_end_pre - p_raw_image->tick_start_pre;
    run_time->npu_proc_time = p_raw_image->tick_end_npu - p_raw_image->tick_start_npu;
    run_time->post_proc_time = p_raw_image->tick_end_post - p_raw_image->tick_start_post;
    run_time->ncpu_req_ipc_ack = p_raw_image->tick_ncpu_img_ack - p_raw_image->tick_ncpu_img_req;
    run_time->ncpu_rslt_ipc_ack = p_raw_image->tick_rslt_got_scpu_ack - p_raw_image->tick_end_post;
    run_time->ncpu_img_req_to_rcv_new_img = p_raw_image->tick_start_pre - p_raw_image->tick_last_img_req;
#endif
    tick_prev_end = p_raw_image->tick_end;
}



int kmdw_model_is_model_loaded(uint32_t model_type)
{
    if (_get_model_info_array_index_by_model_type(model_type) == -1)
        return 0;
    else
        return 1;
}

int32_t kmdw_model_is_model_in_flash(uint32_t model_type)
{
    kmdw_model_fw_info_t *fw_info_ptr = _load_flash_model_info();
    if (NULL == fw_info_ptr)
        return 0;

    if (fw_info_ptr->model_count == 0xFFFFFFFF)
        return 0;
    for (int i = 0; i < fw_info_ptr->model_count; i++) {
        if (fw_info_ptr->models[i].model_type == model_type)
            return 1;
    }

    return 0;
}

uint32_t *kmdw_model_get_all_model_info(bool from_ddr, bool trust_ddr_data)
{
    static uint32_t *s_p_model_id_list = NULL;  //[model_count, id0, id1, id2 ...]

    kmdw_model_fw_info_t *fw_info_ptr = NULL;

    fw_info_ptr = _get_fw_info(from_ddr, trust_ddr_data);

    if (fw_info_ptr) {

        if (NULL == s_p_model_id_list)
            s_p_model_id_list = (uint32_t *)calloc(1+KMDW_MODEL_MAX_MODEL_COUNT, sizeof(uint32_t));

        if (NULL == s_p_model_id_list) {
            err_msg("[ERR] insufficent memory for model id list\n");
        } else {
            int i;
            uint32_t model_id;

            s_p_model_id_list[0] = fw_info_ptr->model_count;
            dbg_msg("%s:\n", __FUNCTION__);
            dbg_msg("Model Count = %d\n", s_p_model_id_list[0]);

            for (i = 0 ; i < s_p_model_id_list[0]; i++) {
                model_id = fw_info_ptr->models[i].model_type;
                dbg_msg("Extract Model ID %d\n", model_id);

                s_p_model_id_list[i+1] = model_id;
            }
        }
        return s_p_model_id_list;
    } else {
        return NULL;
    }

}

uint32_t kmdw_model_get_crc(bool from_ddr, bool trust_ddr_data)
{
    uint32_t ret = 0;
    kmdw_model_fw_info_t *fw_info_ptr;
    kmdw_model_fw_info_ext_t *fw_info_ext_ptr;

    fw_info_ptr = _get_fw_info(from_ddr, trust_ddr_data);
    fw_info_ext_ptr = _get_fw_info_ext_by_fw_info(fw_info_ptr);

    if (fw_info_ext_ptr) {
        ret = fw_info_ext_ptr->model_checksum;
    }

    dbg_msg("%s = 0x%x\n", __FUNCTION__, ret);

    return ret;
}


kmdw_model_fw_info_t *kmdw_model_get_fw_info_buf(void)
{
    return s_fw_info_buf_p;
}

uint32_t kmdw_model_get_model_end_addr(bool from_ddr, bool trust_ddr_data)
{
    uint32_t ret = 0;
    kmdw_model_fw_info_t* fw_info_ptr;
    kmdw_model_fw_info_ext_t* fw_info_ext_ptr = NULL;

    fw_info_ptr = _get_fw_info(from_ddr, trust_ddr_data);
    fw_info_ext_ptr = _get_fw_info_ext_by_fw_info(fw_info_ptr);

    if (fw_info_ext_ptr) {
        ret = fw_info_ext_ptr->model_dram_addr_end;
    }

    dbg_msg("%s = 0x%x\n", __FUNCTION__, ret);

    return ret;
}

#ifndef KDP2_FW
// sorry, due to extern functions

/***************** JPEG enc functions ***********/

void model_jpeg_enc_cfg(uint32_t img_seq, uint32_t jpeg_in_buf_addr, uint32_t jpeg_in_size, uint32_t jpeg_out_buf_addr,
                        uint32_t exec_rlt_addr, uint32_t w, uint32_t h, uint32_t quality, uint32_t fmt, osEventFlagsId_t *pEvt, uint32_t evt_flag)
{
    jpeg_encode_params_t *pEnc;
    scpu_to_ncpu_t *p_out_ipc;

    jpeg_enc_kmdw_ctx.jpeg_enc_evt_id = *pEvt;
    jpeg_enc_kmdw_ctx.jpeg_enc_evt_flag = evt_flag;
    jpeg_enc_kmdw_ctx.p_ipc_out = kmdw_ipc_get_output();
    jpeg_enc_kmdw_ctx.p_ipc_in = kmdw_ipc_get_input();

    p_out_ipc = (scpu_to_ncpu_t *)jpeg_enc_kmdw_ctx.p_ipc_out;

    pEnc = (jpeg_encode_params_t *)p_out_ipc->pExtInParam;
    p_out_ipc->nLenExtInParam = sizeof (jpeg_encode_params_t);

    p_out_ipc->pExtOutRslt = (void *)exec_rlt_addr;
    p_out_ipc->nLenExtOutRslt = sizeof (jpeg_encode_result_t);

    memset((void *)pEnc, 0, sizeof (jpeg_encode_params_t));
    pEnc->frame_seq = img_seq;
    pEnc->rotation = 0;
    pEnc->quality = quality;
    pEnc->src_buf.fmt = (jpeg_fmt_t)fmt;
    pEnc->src_buf.width = w;
    pEnc->src_buf.height = h;
    pEnc->src_buf.yBuf = (uint8_t *)jpeg_in_buf_addr;
    pEnc->src_buf.uvBuf = pEnc->src_buf.yBuf +
                          get_yuv_y_len(pEnc->src_buf.fmt, pEnc->src_buf.width, pEnc->src_buf.height);
    pEnc->dst_buf = jpeg_out_buf_addr;
    pEnc->dst_buf_size = w * h * 2;  // estimated max size

    p_out_ipc->cmd = CMD_JPEG_ENCODE;

}

void model_jpeg_enc_start_run(void)
{
    //trigger ncpu/npu
    kmdw_ipc_trigger_int(CMD_JPEG_ENCODE);
}

jpeg_encode_result_t* model_jpeg_enc_get_rslt()
{
    return (jpeg_encode_result_t*)&jpeg_enc_kmdw_ctx.npu_rslt;
}

/***************** JPEG dec functions ***********/

void model_jpeg_dec_cfg(uint32_t img_seq, uint32_t jpeg_in_buf_addr, uint32_t jpeg_in_size, uint32_t yuv_out_buf_addr,
                        uint32_t exec_rlt_addr, uint32_t w, uint32_t h, uint32_t fmt, osEventFlagsId_t *pEvt, uint32_t evt_flag)
{
    jpeg_decode_params_t *pDec;
    scpu_to_ncpu_t *p_out_ipc;

    jpeg_dec_kmdw_ctx.jpeg_dec_evt_id = *pEvt;
    jpeg_dec_kmdw_ctx.jpeg_dec_evt_flag = evt_flag;
    jpeg_dec_kmdw_ctx.p_ipc_out = kmdw_ipc_get_output();
    jpeg_dec_kmdw_ctx.p_ipc_in = kmdw_ipc_get_input();

    p_out_ipc = (scpu_to_ncpu_t *)jpeg_dec_kmdw_ctx.p_ipc_out;

    pDec = (jpeg_decode_params_t *)p_out_ipc->pExtInParam;
    p_out_ipc->nLenExtInParam = sizeof (jpeg_decode_params_t);

    p_out_ipc->pExtOutRslt = (void *)exec_rlt_addr;
    p_out_ipc->nLenExtOutRslt = sizeof (jpeg_decode_result_t);

    memset((void *)pDec, 0, sizeof (jpeg_decode_params_t));

    pDec->src_buf.buf_addr = jpeg_in_buf_addr;
    pDec->src_buf.buf_filled_len = jpeg_in_size;
    pDec->out_fmt = (jpeg_fmt_t)fmt;
    pDec->width = w;
    pDec->height = h;

    pDec->dst_yuv_buf = yuv_out_buf_addr;
    pDec->dst_buf_size = w * h * 2;  // estimated max size, optional

    pDec->restart = 0;

    /*  gif buf for validation, not for product */
    pDec->ftr.gen_gif = FALSE;
    pDec->gif_buf = NULL;

    pDec->frame_seq = img_seq;

    p_out_ipc->cmd = CMD_JPEG_DECODE;

}

void model_jpeg_dec_start_run(void)
{
    //trigger ncpu/npu
    kmdw_ipc_trigger_int(CMD_JPEG_DECODE);
}

jpeg_decode_result_t* model_jpeg_dec_get_rslt()
{
    return (jpeg_decode_result_t*)&jpeg_dec_kmdw_ctx.npu_rslt;
}

#endif

#if DEBUG

void kmdw_model_dump_model_info(void)
{
    //kdp_model_mgr_t *mgr = kdp_modelmgr();
    struct kdp_model_s *p_modelInfo = 0;
    uint8_t i;

    dbg_msg("Model info Count = %d\n", s_model_data.n_model_count);

    for (i = 0 ; i < s_model_data.n_model_count ; i++) {
        p_modelInfo = &(kmdw_model_data.p_model_info[i]);
        dbg_msg("Model(%2d) model_type(%3d)/version(%5d):\n",
                (i+1),
                p_modelInfo->model_type, p_modelInfo->model_version);

        dbg_msg("input[%x](sz:%d) -> cmd[%x](sz:%d),weight[%x](sz:%d),setup[%x](sz:%d),buf[%x](sz:%d) -> out[%x](sz:%d)\n",
                (i+1),
                p_modelInfo->input_mem_addr, p_modelInfo->input_mem_len,
                p_modelInfo->cmd_mem_addr,   p_modelInfo->cmd_mem_len,
                p_modelInfo->weight_mem_addr,p_modelInfo->weight_mem_len,
                p_modelInfo->setup_mem_addr, p_modelInfo->setup_mem_len,
                p_modelInfo->buf_addr,       p_modelInfo->buf_len,
                p_modelInfo->output_mem_addr,p_modelInfo->output_mem_len);
    }

    return;
}

#endif // DEBUG
