/* Copyright (c) 2020 Kneron, Inc. All Rights Reserved.
 *
 * The information contained herein is property of Kneron, Inc.
 * Terms and conditions of usage are described in detail in Kneron
 * STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information.
 * NO WARRANTY of ANY KIND is provided. This heading must NOT be removed
 * from the file.
 */
#include <stdint.h>
#include "kdev_status.h"
#include "kdrv_uart.h"
#include "kmdw_sbt.h"
#include "kmdw_console.h"
#include "kmdw_memxfer.h"
#include "kdrv_aes.h"
#include "kdrv_hash.h"
#include "kmdw_dfu.h"
#include "membase.h"
#include "string.h"
#include "dbg.h"
#define SBT_DBG
#ifdef SBT_DBG
#define sbt_msg(fmt, ...) kmdw_printf("[SBT] " fmt, ##__VA_ARGS__)
#else
#define flash_msg(fmt, ...)
#endif
#define sbt_err_msg(fmt, ...) kmdw_printf("[SBT] " fmt, ##__VA_ARGS__)

#define BIN_MAGIC_NUM           0x03020100
#define BIN_HEADER_SIZE         16
#define BIN_HEADER_ENC_SIZE     48
#define BIN_HEADER_TSIZE        (BIN_HEADER_SIZE + BIN_HEADER_ENC_SIZE)
typedef struct {
    uint32_t    magic_num;
    uint32_t    size;
    uint32_t    reserved;
    uint32_t    checksum;
} bin_header_t;

typedef struct {
    uint32_t    auth_num;
    uint32_t    size;
    uint8_t     hash[32];
    uint64_t    reserved;
} bin_header_enc_t;

#define EFUSE_AUTH_NUM_SIZE     4
#define EFUSE_KEY_SIZE          32
#define EFUSE_IV_SIZE           16
typedef struct {
    uint32_t    auth_num;
    uint8_t     key[EFUSE_KEY_SIZE];
    uint8_t     iv[EFUSE_IV_SIZE];
} efuse_key_info_t;

efuse_key_info_t key_info;
bin_header_t header;
bin_header_enc_t header_enc;
efuse_key_info_t efuse_info;

extern kdev_status_t kdev_flash_readdata(uint32_t addr, void *data, uint32_t cnt);
extern bool kdrv_efuse_get_enc_boot_flag(void);
extern void kdrv_efuse_get_auth(void *val, uint8_t size);
extern void kdrv_efuse_get_key(void *val, uint8_t size);
extern void kdrv_efuse_get_iv(void *val, uint8_t size);
/**
 * @brief kmdw_sbt_get_boot_flag() get efuse secure boot flag
 *
 * @return bool             secure boot flag 1:enable, 0: disable
 */
bool kmdw_sbt_get_boot_flag(void)
{
    return(kdrv_efuse_get_enc_boot_flag());
}

/*
 * @brief kmdw_sbt_efuse_read() read key information from efuse.
 */
static void kmdw_sbt_efuse_read(void)
{
    kdrv_efuse_get_auth((void*)&efuse_info.auth_num, EFUSE_AUTH_NUM_SIZE);
    kdrv_efuse_get_key((void*)&efuse_info.key, EFUSE_KEY_SIZE);
    kdrv_efuse_get_iv((void*)&efuse_info.iv, EFUSE_IV_SIZE);
}


/*
 * @brief kmdw_sbt_flash_header_read()
 *
 * @return  kdrv_status_t
 */
kdrv_status_t kmdw_sbt_flash_header_read(uint32_t flash_addr, block_t *key_blk, block_t *iv_blk)
{
    kdrv_status_t ret;
    uint8_t rbuf[BIN_HEADER_ENC_SIZE];
    block_t out_blk = block_t_convert((const volatile void *)&header_enc, sizeof(header_enc));
    block_t r_blk = block_t_convert((const volatile void *)rbuf, sizeof(rbuf));
    /* get normal header */
    if((ret = (kdrv_status_t)kdev_flash_readdata(flash_addr, (void*)&header, sizeof(header)))
                                                          != KDRV_STATUS_OK)
        return (kdrv_status_t)ret;
    if (header.magic_num != BIN_MAGIC_NUM) {
        sbt_err_msg("ERR: flash header format error\n");
        return KDRV_STATUS_ERROR;
    }

    /* get encrypted header */
    if((ret = (kdrv_status_t)kdev_flash_readdata(flash_addr + BIN_HEADER_SIZE
                                    , (void*)&rbuf, sizeof(rbuf)))
                                    != KDRV_STATUS_OK) {
        return KDRV_STATUS_ERROR;
    }
    ret = kdrv_aes_cbc_decrypt(key_blk, iv_blk, &r_blk, &out_blk); 
    if (ret != KDRV_STATUS_OK) {        
        sbt_err_msg("ERR: decoder fail\r\n");
        return ret;
    }
    return KDRV_STATUS_OK;
}

/*
 * @brief kmdw_sbt_flash_fw_loader(), loader encryption firmware
 *
 * @param[in]   flash_addr          flash address of firmware
 * @param[in]   cpu                 cpu type, ref @fw_type_e
 * @param[in]   swap_buf            swap buffer for decrypt process
 * @param[in]   swap_buf_size       swap buffer size
 * @return kdrv_status_t
 */
kdrv_status_t kmdw_sbt_flash_fw_loader(uint32_t flash_addr, fw_type_e cpu, uint32_t *swap_buf, uint32_t swap_buf_size)
{
    uint32_t i, cp_loop, cp_remainder;
    uint32_t cpu_addr;
    uint8_t iv[EFUSE_IV_SIZE];
    kdrv_status_t ret;
    kmdw_sbt_efuse_read();
    kdrv_aes_initialize();
    memcpy((void*)iv, (void*)efuse_info.iv, sizeof(iv));
    block_t key_blk = block_t_convert(efuse_info.key, sizeof(efuse_info.key));
    block_t iv_blk = block_t_convert(iv,  sizeof(iv));    
    /* read and check scpu flash header and header format */
    if ((ret = kmdw_sbt_flash_header_read(flash_addr, &key_blk, &iv_blk)) != KDRV_STATUS_OK)
        return ret;
    if (header_enc.auth_num != efuse_info.auth_num) {
        if (cpu == FW_SCPU)
            sbt_err_msg("SCPU auth value fail\r\n");
        else
            sbt_err_msg("NCPU auth value fail\r\n");
        return KDRV_STATUS_ERROR;
    }
    //DDR_MEM_PHY_BASE, NCPU_START_ADDRESS
    cpu_addr = cpu == FW_SCPU ? SCPU_START_ADDRESS : DDR_MEM_PHY_BASE;
    memset((void*)cpu_addr, 0, header_enc.size);
    
    cp_loop = (uint32_t)(header_enc.size + swap_buf_size -1) / swap_buf_size;
    cp_remainder = header_enc.size % swap_buf_size;
    //sbt_msg("cp loop=%x, cp_remainder=%x (len=%x, swap_buf_size=%x)\r\n", 
    //            cp_loop, cp_remainder, header_enc.size, swap_buf_size);
    block_t enbuf_blk = block_t_convert((const volatile void *) swap_buf, swap_buf_size);
    for (i = 0; i < cp_loop; i++) {
        /* store new iv */
        if (i != 0) {
            memcpy((void*) iv, swap_buf + ((swap_buf_size - 16) / 4) , sizeof(iv));
        }

        /* modified decrypt output address */
        block_t obuf_blk = block_t_convert(
            (const volatile void *)(cpu_addr + i * swap_buf_size), swap_buf_size);
        ret = (kdrv_status_t)kdev_flash_readdata(flash_addr + BIN_HEADER_TSIZE 
                + i * swap_buf_size, (void*)swap_buf, swap_buf_size);
        if (ret != KDRV_STATUS_OK)
            return KDRV_STATUS_ERROR;
        ret = kdrv_aes_cbc_decrypt(&key_blk, &iv_blk, &enbuf_blk, &obuf_blk);    
        if (ret != KDRV_STATUS_OK) {        
            sbt_err_msg("SCPU decoder fail\r\n");
            return ret;
        }
    }
    if (cp_remainder != 0) {
        block_t enbuf_blk = block_t_convert((const volatile void *) swap_buf, cp_remainder);
        block_t obuf_blk = block_t_convert((const volatile void *)
                (cpu_addr + cp_loop * swap_buf_size), cp_remainder);
        if((ret = (kdrv_status_t)kdev_flash_readdata
                (flash_addr + BIN_HEADER_TSIZE + cp_loop * swap_buf_size, (void*)swap_buf, cp_remainder)) != KDRV_STATUS_OK)
            return KDRV_STATUS_ERROR;
        ret = kdrv_aes_cbc_decrypt(&key_blk, &iv_blk, &enbuf_blk, &obuf_blk);    
        if (ret != KDRV_STATUS_OK) {        
            sbt_err_msg("decoder fail\r\n");
            return ret;
        }
    }
    // get hash value
    kdrv_hash_initialize();
    block_t msg_blk = block_t_convert(
            (const volatile void *)cpu_addr, header_enc.size);
    uint8_t out[32] = {0};
    ret = kdrv_hash_blk(e_SHA256, msg_blk, BLK_LITARRAY(out));
    if (memcmp((void*)out, (const void*)header_enc.hash, sizeof(out)) != 0) {
        if (cpu == FW_SCPU)
            sbt_msg("SCPU hash compare fail\r\n");
        else
            sbt_msg("NCPU hash compare fail\r\n");
        return KDRV_STATUS_ERROR;
    }
    if (cpu == FW_NCPU) {
        memcpy((void *)NiRAM_MEM_BASE, (void *)DDR_MEM_PHY_BASE, NiRAM_MEM_SIZE);
    }
    return KDRV_STATUS_OK;
}


