/********************************************************************
 * Copyright (c) 2020 Kneron, Inc. All Rights Reserved.
 *
 * The information contained herein is property of Kneron, Inc.
 * Terms and conditions of usage are described in detail in Kneron
 * STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information.
 * NO WARRANTY of ANY KIND is provided. This heading must NOT be removed
 * from the file.
 ********************************************************************/

/**@addtogroup  KDRV_USBD3
 * @{
 * @brief       Kneron USB3 device mode driver
 * @version     v1.0
 * @copyright   Copyright (C) 2020 Kneron, Inc. All rights reserved.
 */

#ifndef __KDRV_USBD3_H__
#define __KDRV_USBD3_H__

#include "kdrv_status.h"

#define MAX_USBD3_ENDPOINT 4 /**< maximum number of endpoint descriptor */

/** @brief Structure of 8-byte setup packet */
typedef struct __attribute__((__packed__))
{
    uint8_t bmRequestType;
    uint8_t bRequest;
    uint16_t wValue;
    uint16_t wIndex;
    uint16_t wLength;
} kdrv_usbd3_setup_packet_t;

/** @brief Structure of Device descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t bcdUSB;
    uint8_t bDeviceClass;
    uint8_t bDeviceSubClass;
    uint8_t bDeviceProtocol;
    uint8_t bMaxPacketSize0;
    uint16_t idVendor;
    uint16_t idProduct;
    uint16_t bcdDevice;
    uint8_t iManufacturer;
    uint8_t iProduct;
    uint8_t iSerialNumber;
    uint8_t bNumConfigurations;
} kdrv_usbd3_device_descriptor_t;

/** @brief Structure of Configuration descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t wTotalLength;
    uint8_t bNumInterfaces;
    uint8_t bConfigurationValue;
    uint8_t iConfiguration;
    uint8_t bmAttributes;
    uint8_t MaxPower;
} kdrv_usbd3_config_descriptor_t;

/** @brief Structure of Interface descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bInterfaceNumber;
    uint8_t bAlternateSetting;
    uint8_t bNumEndpoints;
    uint8_t bInterfaceClass;
    uint8_t bInterfaceSubClass;
    uint8_t bInterfaceProtocol;
    uint8_t iInterface;
} kdrv_usbd3_interface_descriptor_t;

/** @brief Structure of Endpoint descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bEndpointAddress;
    uint8_t bmAttributes;
    uint16_t wMaxPacketSize;
    uint8_t bInterval;
} kdrv_usbd3_endpoint_descriptor_t;

/** @brief Structure of SS Endpoint companion descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bMaxBurst;
    uint8_t bmAttributes;
    uint16_t wBytesPerInterval;
} kdrv_usbd3_endpoint_companion_descriptor_t;

/** @brief Structure of device qualifier descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t bcdUSB;
    uint8_t bDeviceClass;
    uint8_t bDeviceSubClass;
    uint8_t bDeviceProtocol;
    uint8_t bMaxPacketSize0;
    uint8_t bNumConfigurations;
    uint8_t bReserved;
} kdrv_usbd3_device_qualifier_descriptor_t;

/** @brief Structure of all High-Speed descriptors combination  */
typedef struct __attribute__((__packed__))
{
    kdrv_usbd3_device_descriptor_t *dev_descp;
    kdrv_usbd3_config_descriptor_t *config_descp;  // we support only 1 config
    kdrv_usbd3_interface_descriptor_t *intf_descp; // we support only 1 interface
    kdrv_usbd3_endpoint_descriptor_t *enp_descp[MAX_USBD3_ENDPOINT];
    kdrv_usbd3_device_qualifier_descriptor_t *qual_descp; // only for HS, optional
} kdrv_usbd3_HS_descriptors_t;

/** @brief Structure of all Super-Speed descriptors combination  */
typedef struct __attribute__((__packed__))
{
    kdrv_usbd3_device_descriptor_t *dev_descp;
    kdrv_usbd3_config_descriptor_t *config_descp;  // we support only 1 config
    kdrv_usbd3_interface_descriptor_t *intf_descp; // we support only 1 interface
    kdrv_usbd3_endpoint_descriptor_t *enp_descp[MAX_USBD3_ENDPOINT];
    kdrv_usbd3_endpoint_companion_descriptor_t *enp_cmpn_descp[MAX_USBD3_ENDPOINT];
} kdrv_usbd3_SS_descriptors_t;

/** Device string descriptor */
typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint8_t bString[32];
} kdrv_usbd3_prd_string_descriptor_t;

typedef struct __attribute__((__packed__))
{
    uint8_t bLength;
    uint8_t bDescriptorType;
    uint16_t bLanguageID;
    kdrv_usbd3_prd_string_descriptor_t *desc[3];
} kdrv_usbd3_string_descriptor_t;
/** @brief Enumerations of all usb3 speed  */
typedef enum
{
    USBD3_NO_LINK_SPEED = 0,        /**< Enum 0, USB3 no link speed */
    USBD3_HIGH_SPEED,               /**< Enum 1, USB3 high speed */
    USBD3_SUPER_SPEED               /**< Enum 2, USB3 super speed */
} kdrv_usbd3_speed_t;

/** @brief Enumerations of all usb3 link status */
typedef enum
{
    USBD3_STATUS_DISCONNECTED = 0x0,    /**< Enum 0, USB3 status disconnection */
    USBD3_STATUS_CONFIGURED,            /**< Enum 1, USB3 status connections */
} kdrv_usbd3_link_status_t;

typedef void (*kdrv_usbd3_link_status_callback_t)(kdrv_usbd3_link_status_t link_status);
typedef bool (*kdrv_usbd3_user_control_callback_t)(kdrv_usbd3_setup_packet_t *setup);

/**
 * @brief           USB3 device mode driver initialization
 * @param[in]       hs_descs              user created HS device descriptor, this must be kept during device enumeration, see @ref kdrv_usbd3_HS_descriptors_t
 * @param[in]       ss_descs              user created SS device descriptor, this must be kept during device enumeration, see @ref kdrv_usbd3_SS_descriptors_t
 * @param[in]       status_isr_cb        USBD event callback function, can be NULL, see @ref kdrv_usbd3_link_status_callback_t
 * @param[in]       usr_cx_isr_cb       see @ref kdrv_usbd3_user_control_callback_t

 *
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_initialize(
    kdrv_usbd3_HS_descriptors_t *hs_descs,
    kdrv_usbd3_SS_descriptors_t *ss_descs,
    kdrv_usbd3_string_descriptor_t *dev_str_desc,
    kdrv_usbd3_link_status_callback_t status_isr_cb,
    kdrv_usbd3_user_control_callback_t usr_cx_isr_cb);

/**
 * @brief          USB3 device mode driver de-initialization
 *
 * @param[in]       N/A
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_uninitialize(void);

/**
 * @brief           set enable/disabale of USB device mode, host can enumerate this device only if device is enabled
 *
 * @param[in]       enable                true to enable, false to disable
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_set_enable(bool enable);

/**
 * @brief          Reset USB3 device
 *
 * @param[in]       N/A
 * @note            this cannot be called in ISR context
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_reset_device(void);

/**
 * @brief          Reset USB3 endpoint
 *
 * @param[in]       N/A
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_reset_endpoint(uint8_t endpoint);

/**
 * @brief          Get USB3 link status
 *
 * @param[in]       N/A
 * @return          kdrv_usbd3_link_status_t         see @ref kdrv_usbd3_link_status_t
 */
kdrv_usbd3_link_status_t kdrv_usbd3_get_link_status(void);

/**
 * @brief          Get USB3 link speed
 *
 * @param[in]       N/A
 * @return          kdrv_usbd3_speed_t         see @ref kdrv_usbd3_speed_t
 */
kdrv_usbd3_speed_t kdrv_usbd3_get_link_speed(void);

/**
 * @brief           Bulk-OUT transfser, receive data from the host through a bulk-out endpoint in blocking mode
 *
 * @param[in]       endpoint              a bulk-out endpoint address, should be the value from bEndpointAddress
 * @param[out]      buf                   buffer for receiving data
 * @param[in,out]   blen                  buffer length for input, actual transfered length for output
 * @param[in]       timeout_ms            timeout in millisecond
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_bulk_receive(uint8_t endpoint, void *buf, uint32_t *blen, uint32_t timeout_ms);

/**
 * @brief          Bulk-OUT transfser, receive zip data from the host through a bulk-out endpoint in blocking mode
 *
 * @param[in]    endpoint                 a bulk-out endpoint address, should be the value from bEndpointAddress
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_bulk_receive_zlp(uint8_t endpoint);

/**
 * @brief           Bulk-IN transfser, send data to the host through a bulk-in endpoint in blocking mode
 *
 * @param[in]       endpoint              a bulk-in endpoint address, should be the value from bEndpointAddress
 * @param[out]      buf                   buffer for sending data
 * @param[in]       txlen                 data length to be sent
 * @param[in]       timeout_ms            timeout in millisecond
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_bulk_send(uint8_t endpoint, void *buf, uint32_t txlen, uint32_t timeout_ms);

/**
 * @brief           Interrupt transfser
 *
 * @param[in]       endpoint              a bulk-in endpoint address, should be the value from bEndpointAddress
 * @param[out]      buf                   buffer for sending data
 * @param[in]       txlen                 data length to be sent
 * @param[in]       timeout_ms            timeout in millisecond
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_usbd3_interrupt_send(uint32_t endpoint, void *buf, uint32_t txLen, uint32_t timeout_ms);

#endif
/** @}*/

