/* #include <iostream> */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>


typedef unsigned char           u8;
typedef unsigned short          u16;
typedef unsigned int            u32;
typedef unsigned long long      u64;


#define SCPU_IMAGE_SIZE             0x20000    // 128KB
#define NCPU_IMAGE_SIZE             0x200000   // 2MB


#define FW_INFO_SIZE       0x20000               //NOR: 4KB, NAND:128KB
#define MAX_ALL_MODEL_SIZE (100 * 1024 * 1024)   // 100MB

/* Calculate the 32-bit checksum of object using 32-bit size block
** If address is not 32-bit aligned, use the fraction of the block w/ zero's
** replacing excluded bytes.  (Note: we use Little Endian)
*/
u32 kdp_gen_sum32(u8 *data, u32 size)
{
    u32 sum, *ddr, i = (uintptr_t)data & 0x03;
    ddr = (u32*)((uintptr_t)data & 0xFFFFFFFFFFFFFFFC);  // point to the first 32-bit aligned block

    if (i) { // starting address misaligned ?
        size = size + i - 4;
        sum = *ddr;
        sum >>= i * 8;
        sum <<= i * 8;
        ddr++;
    }
    else
        sum = 0;
    for (i = 0; i < (size & 0xFFFFFFFC); i += 4) {
        sum += *ddr;
        ddr++;
    }
    i = size & 3;
    if (i)  // ending address misaligned ?
    {
        size = *ddr;
        size <<= (4 - i) * 8;
        size >>= (4 - i) * 8;
        sum += size;
    }
    return(sum);
}

int gen_dfu_scpu(char * in_file, char * out_file)
{
    FILE *pfIn, *pfOut;
    u32 sum32;

    pfIn = fopen(in_file, "rb");
    if (pfIn == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, in_file);
        return -1;
    }

    pfOut = fopen(out_file, "wb");
    if (pfOut == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, out_file);
        return -1;
    }

    fseek(pfIn, 0, SEEK_END);
    int inFileLen = ftell(pfIn);
    if (inFileLen > SCPU_IMAGE_SIZE)
    {
        printf("Error: intput SCPU file is larger than 90112 bytes\n");
        return -1;
    }
    fseek(pfIn, 0, SEEK_SET);

    u8 * buf = (u8 *)malloc(SCPU_IMAGE_SIZE);
    if (buf == NULL)
    {
        printf("Error: failed to alloc memory\n");
        return -1;
    }

    memset(buf, 0, SCPU_IMAGE_SIZE);

    long result = fread(buf, 1, inFileLen, pfIn);
    if (result != inFileLen)
    {
        printf("Error @ %s:%d: file read less bytes\n", __func__, __LINE__);
        return 0;
    }

    fclose(pfIn);

    memset(buf + inFileLen, 0, SCPU_IMAGE_SIZE - inFileLen);

    sum32 = kdp_gen_sum32(buf, SCPU_IMAGE_SIZE - 4);

    *(u32 *)(buf + SCPU_IMAGE_SIZE - 4) = sum32;

    fwrite(buf, SCPU_IMAGE_SIZE, 1, pfOut);
    fclose(pfOut);
    return 0;
}

int gen_dfu_ncpu(char * in_file, char * out_file)
{
    FILE *pfIn, *pfOut;
    u32 sum32;

    pfIn = fopen(in_file, "rb");
    if (pfIn == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, in_file);
        return -1;
    }

    pfOut = fopen(out_file, "wb");
    if (pfOut == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, out_file);
        return -1;
    }

    fseek(pfIn, 0, SEEK_END);
    int inFileLen = ftell(pfIn);
    if (inFileLen > NCPU_IMAGE_SIZE)
    {
        printf("Error: intput NCPU file is larger than 65536 bytes\n");
        return -1;
    }
    fseek(pfIn, 0, SEEK_SET);

    u8 * buf = (u8 *)malloc(NCPU_IMAGE_SIZE);
    if (buf == NULL)
    {
        printf("Error: failed to alloc memory\n");
        return -1;
    }

    memset(buf, 0, NCPU_IMAGE_SIZE);

    long result = fread(buf, 1, inFileLen, pfIn);
    if (result != inFileLen)
    {
        printf("Error @ %s:%d: file read less bytes\n", __func__, __LINE__);
        return 0;
    }

    fclose(pfIn);

    memset(buf + inFileLen, 0, NCPU_IMAGE_SIZE - inFileLen);

    sum32 = kdp_gen_sum32(buf, NCPU_IMAGE_SIZE - 4);
    *(u32 *)(buf + NCPU_IMAGE_SIZE - 4) = sum32;

    fwrite(buf, NCPU_IMAGE_SIZE, 1, pfOut);
    fclose(pfOut);
    return 0;
}

int gen_dfu_model(char * fw_info_file, char * all_model_file, char * out_file)
{
    FILE *pfFW, *pfModel, *pfOut;
    u32 sum32;
    u32 total_size;

    pfFW = fopen(fw_info_file, "rb");
    if (pfFW == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, fw_info_file);
        return -1;
    }

    pfModel = fopen(all_model_file, "rb");
    if (pfModel == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, all_model_file);
        return -1;
    }

    pfOut = fopen(out_file, "wb");
    if (pfOut == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, out_file);
        return -1;
    }

    fseek(pfFW, 0, SEEK_END);
    u32 FwFileLen = ftell(pfFW);
    if (FwFileLen > FW_INFO_SIZE)
    {
        printf("Error: intput fw_info file is larger than 232 bytes\n");
        return -1;
    }
    fseek(pfFW, 0, SEEK_SET);

    fseek(pfModel, 0, SEEK_END);
    u32 ModelFileLen = ftell(pfModel);
    if (ModelFileLen > MAX_ALL_MODEL_SIZE)
    {
        printf("Error: intput all_model_info file is larger than %d bytes\n", MAX_ALL_MODEL_SIZE);
        return -1;
    }
    fseek(pfModel, 0, SEEK_SET);

    total_size = FW_INFO_SIZE + ModelFileLen + 4;   // 4 bytes for sum32 word

    if ((total_size % 4) != 0)
    {
        total_size = ((total_size + 3) / 4) * 4;
    }

    u8 * buf = (u8 *)malloc(total_size);
    if (buf == NULL)
    {
        printf("Error: failed to alloc memory\n");
        return -1;
    }

    memset(buf, 0, total_size);

    long result = fread(buf, 1, FwFileLen, pfFW);
    if (result != FwFileLen)
    {
        printf("Error @ %s:%d: file read less bytes\n", __func__, __LINE__);
        return 0;
    }
    fclose(pfFW);
    memset(buf + FwFileLen, 0xff, FW_INFO_SIZE - FwFileLen);


    u8 * pModel = buf + FW_INFO_SIZE;
    result = fread(pModel, 1, ModelFileLen, pfModel);
    if (result != ModelFileLen)
    {
        printf("Error @ %s:%d: file read less bytes\n", __func__, __LINE__);
        return 0;
    }
    fclose(pfModel);

    sum32 = kdp_gen_sum32(buf, total_size - 4);

    *(u32 *)(buf + total_size - 4) = sum32;

    fwrite(buf, total_size, 1, pfOut);
    fclose(pfOut);
    free(buf);

    return 0;
}

/*
convert the SPL binary file to the fixed size 8K for flash convenience

*/

int gen_dfu_spl(char * in_file, char * out_file)
{
    FILE *pfIn, *pfOut;

    pfIn = fopen(in_file, "rb");
    if (pfIn == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, in_file);
        return -1;
    }

    pfOut = fopen(out_file, "wb");
    if (pfOut == NULL)
    {
        printf("Error @ %s:%d: failed on open file %s\n", __func__, __LINE__, out_file);
        return -1;
    }

    fseek(pfIn, 0, SEEK_END);
    int inFileLen = ftell(pfIn);
    if (inFileLen > 8192)
    {
        printf("Error: intput SPL file is larger than 8192 bytes\n");
        return -1;
    }
    fseek(pfIn, 0, SEEK_SET);

    u8 * buf = (u8 *)malloc(8192);
    if (buf == NULL)
    {
        printf("Error: failed to alloc memory\n");
        return -1;
    }

    memset(buf, 0, 8192);

    long result = fread(buf, 1, inFileLen, pfIn);
    if (result != inFileLen)
    {
        printf("Error @ %s:%d: file read less bytes\n", __func__, __LINE__);
        return 0;
    }

    fclose(pfIn);

    memset(buf + inFileLen, 0, 8192 - inFileLen);

    fwrite(buf, 8192, 1, pfOut);
    fclose(pfOut);
    return 0;
}


/* usage: generate_dfu_binary -scpu scpu_in_file scpu_out_file
or generate_dfu_binary -ncpu ncpu_in_file ncpu_out_file
or generate_dfu_binary -model fw_info_file_in  all_model_file_in model_out_file
*/

int main(int argc, char* argv[])
{

    char *in_file, *out_file, *fw_info_file, *all_model_file;

    if ((argc != 4) && (argc != 5))
    {
        printf("This utility converts scpu/ncpu/model binary to dfu format output\n");
        printf("usage 1: generate_dfu_binary -scpu scpu_in_file scpu_out_file\n");
        printf("usage 2: generate_dfu_binary -ncpu ncpu_in_file ncpu_out_file\n");
        printf("usage 3: generate_dfu_binary -model fw_info_file_in  all_model_file_in model_out_file\n");
        printf("usage 4: generate_dfu_binary -spl spl_file_in  spl_out_file\n");
        return 0;
    }

    if (strcmp(argv[1], "-scpu") == 0)
    {
        in_file = argv[2];
        out_file = argv[3];
        gen_dfu_scpu(in_file, out_file);
        return 0;
    }

    if (strcmp(argv[1], "-ncpu") == 0)
    {
        in_file = argv[2];
        out_file = argv[3];
        gen_dfu_ncpu(in_file, out_file);
        return 0;
    }

    if (strcmp(argv[1], "-model") == 0)
    {
        fw_info_file = argv[2];
        all_model_file = argv[3];
        out_file = argv[4];
        gen_dfu_model(fw_info_file, all_model_file, out_file);
        return 0;
    }

    if (strcmp(argv[1], "-spl") == 0)
    {
        in_file = argv[2];
        out_file = argv[3];
        gen_dfu_spl(in_file, out_file);
        return 0;
    }
}
